# -*- coding: utf-8 -*-
#
# “My Hovercraft is Full of Eels” translations
#
# Taken from
#   "Omniglot - writing systems and languages of the world".
#   Copyright (c) 1998-2018 Simon Ager. All rights reserved.
#
# Redistributed for non-commercial purposes under the copyright notice
# offered by Simon Ager on the omniglot.com web site:
#
#   “If you would like to use the material for personal, educational or
#   non-commercial purposes you're welcome to do so. Don't forget to mention
#   where it came from.”
#
# For the original compilation of translations,
# see <http://www.omniglot.com/language/phrases/hovercraft.htm>.
#
# For the copyright notice allowing distribution for non-commercial purposes,
# see <http://www.omniglot.com/copyright.htm>.
#
#
# Separated as file with nonfree material, because the author of TypeAtlas is
# not a lawyer, and cannot make interpretations of copyright law beyond
# the established license or notices, presented above. This separation itself
# does not constitute a claim that the file is indeed non-free, should it be
# used in jurisdictions where data or a database of user-submitted phrases
# (in the manner used by TypeAtlas) *hypotethically* might be
# distributable under terms satisfying The Free Software Definition. For any use
# of this file beyond the scope of the notice from Simon Ager above, consult a
# lawyer.
#

HOVERCRAFT_ENGLISH = 'My hovercraft is full of eels'
HOVERCRAFT_TRANSLATION_SOURCE = 'http://www.omniglot.com/language/phrases/hovercraft.htm'
HOVERCRAFT_NOTE = 'The phrase originated in Monty Python'

hovercraft = {
    'en': HOVERCRAFT_ENGLISH,
    'af': 'My skeertuig is vol palings',
    'sq-alt': 'Anija jêm ãsht plot mê ngjala',
    'sq': 'Automjeti im është plot me ngjala',
    'am': 'የኔ ማንዣበቢያ መኪና በእባቦች ተሞልቷል',
    'ar': 'حَوّامتي مُمْتِلئة بِأَنْقَلَيْسون',
    'an': "O mío aeroeslisador ye pleno d'anguilas",
    'hy': 'Իմ օդաթիռը լի է օձաձկերով',
    'hyw': ' Իմ սաւառնակս օձաձկներով լեցուն է:',
    'as': "মোৰ হ'ভাৰক্ৰাফ্টটো ঈল মাছেৰে ভৰি গৈছে",
    'az': 'Hoverkraftimin içi ilan balıǧı ilə doludur',
    'eu': 'Nire aerolabaingailua aingirez beteta dago',
    'be': 'Маё судна на паветранай падушцы поўна вуграмі',
    'bn': 'আমার হভারক্রাফ্ট কুঁচে মাছ-এ ভরা হয়ে গেছে',
    'bs': 'Moja lebdjelica je puna jegulja',
    'br': 'Leun gant sili eo ma dourruzer',
    'bg': 'Корабът ми на въздушна възглавница е пълен със змиорки',
    'my': 'ကျွန်တော်ရဲ့ လေစီးယာဉ်မှာ ငါးရှင့်တွေအပြည့်ရှိနေပါတယ်။',
    'ca': "El meu aerolliscador està ple d'anguiles",
    'ch': 'Bula titugi eyo-ku hovercraft',
    'ch-gu': 'Bula titugi eyo-ku hovercraft',
    'ch-mp': 'Bula hasuli gi halom hovercraft-hu',
    'zh-yue': '(Cantonese) 我隻氣墊船裝滿晒鱔',
    'zh': '我的氣墊船裝滿了鱔魚 [我的气垫船装满了鳝鱼]',
    'zh-tw': '我的氣墊船充滿了鰻魚',
    'kw': 'Leun a sylli yw ow skath bargesi',
    'co': "U me battellu hè carcu d'anguili",
    'hr': 'Moja je lebdjelica puna jegulja',
    'cs': 'Moje vznášedlo je plné úhořů',
    'da': 'Mit luftpudefartøj er fyldt med ål',
    'nl': 'Mijn luchtkussenboot zit vol paling',
    'et': 'Mu hõljuk on angerjaid täis',
    'fo': "Luftpútufarið hjá mær yður í álli",
    'fj': 'Na noqu hovercraft e sinai tu vei na duna',
    'fi': 'Ilmatyynyalukseni on täynnä ankeriaita',
    'fr': "Mon aéroglisseur est plein d'anguilles",
    'fy': 'Myn hovercraft sit fol mei iel',
    'frr': 'Min luftdümpetbüüdj as ful ma äil',
    'gl': 'O meu aerodeslizador esta cheo de anguías',
    'ka': 'ჩემი ხომალდი საჰაერო ბალიშზე სავსეა გველთევზებით',
    'de': 'Mein Luftkissenfahrzeug ist voller Aale',
    'el-ancient': 'Τὸ ἐμὸν αερόστρωμνον ἐγχελείων πλῆρές ἐστιν',
    'el': 'Το αερόστρωμνο μου ειναι γεμάτο χέλια',
    'gu': 'મારું હોવેરક્રાફ્ટ ઈલ માછલીઓ થી ભરાઈ ગયું છે',
    'ha': 'Jirgina a cike yake da bano',
    'haw': 'Pihaʻū oʻu mokukauaheahe i nā puhi',
    'he': 'הרחפת שלי מלאה בצלופחים',
    'hi': 'मेरी मँडराने वाली नाव सर्पमीनों से भरी हैं',
    'hu': 'A légpárnás hajóm tele van angolnákkal',
    'is': 'Svifnökkvinn minn er fullur af álum',
    'ig': 'Hovercraft m juputara azu arira',
    'smn': 'Muu áimukuáddákäärbis lii tievâ aŋgeriaid',
    'id': 'Hovercraft saya penuh dengan belut',
    'iu': 'ᐅᒥᐊᕐᔫᑉ ᐳᓪᓕᓕᒫᐸᒐ ᑕᑦᑕᕐᓂᖅ ᐊᒻᒪᔭᖅ',
    'ga': "Tá m'árthach foluaineach lán d'eascanna",
    'it': 'Il mio aeroscivolante è pieno di anguille',
    'ja': '私のホバークラフトは鰻でいっぱいです',
    'kn': 'ನನ್ನ ಯಾಂತ್ರಿಕದೋಣಿ ಹಾವುಮೀನುಗಳಿಂದ ತುಂಬಿದೆ',
    'ks': 'Myon hovercraaft chhu eelav seeth bharith',
    'kk': 'Менің әуе негіз кемесi жыланбалықпен толтырылған',
    'km': 'សុទ្ធ​តែ​អន្ទង់​ពេញ​ទូក​ហោះ​យើង',
    'ki': 'E onrake au kaibuke n te rabono',
    'ko': '내 호버크라프트는 장어로 가득 차 있어요',
    'ku-alt': 'Hoverkrafta mi(n) tijé marmasî e',
    'ku': 'Seyarey balindîyekem pirr ye li marmasî',
    'ky': 'Менин аба кемем курт балыктар менен толуп турат',
    'lo': 'ມີປາໄຫຼເຕັມຢູ່ໃນເຮືອພັດລົມຂອງຂ້ອຍ',
    'la-alt': 'Mea navis volitans anguillis plena est',
    'la': 'Navis volitans mihi anguillis plena est',
    'lv': 'Mans transportlīdzeklis uz gaisa spilvena ir pilns ar zušiem',
    'lt': 'Lithuanian  Mano laivas su oro pagalve pilnas ungurių',
    'nds': 'Mien Lüchtpöukaan ist vull von sük aalen',
    'lb': 'Mäi Loftkësseboot ass voller Éilen',
    'mk': 'Моето летачко возило е полно со јагули',
    'ms': 'Hoverkraf saya penuh dengan belut',
    'ml': 'എന്‍റെ പറക്കും-പേടകം നിറയെ വ്ളാങ്കുകളാണ്',
    'mt': 'Il-hovercraft tiegħi hu mimli sallur',
    'gv-alt': 'Ta my haagh crowal lane dy astanyn',
    'gv': 'Ta my lhong chrowal lane dy astanyn',
    'mi': 'Kī tōnu taku waka topaki i te tuna',
    'mr': 'माझी होडी मासळयांनी भरली आहे',
    'mn': 'Миний хөвөгч онгоц могой загасаар дүүрсэн байна',
    'nd': 'Isikepe sami sigcwele umzowa',
    'se': 'Mu áibmofatnasis lea dievva ággarasaid',
    'no-alt': 'Det er fullt av ål i luftputebåten min',
    'no-alt': 'Luftputebåten min er full av ål',
    'no': 'Luftputefartøyet mitt er fullt av ål',
    'oc': "Mon aerolisador es plen d'anguilas",
    'or': 'ମୋ ହୋବର୍କ୍ରାଫ୍ଟ ରେ ଇଲ୍ ଭର୍ତି ହେଇ ଜାଇଛି।',
    'om': 'Koonkalata baariisu ko bofa bishaani gutu',
    'fa': 'هاورکرافت من پر مارماهى است',
    'pl': 'Mój poduszkowiec jest pełen węgorzy',
    'pt': 'O meu hovercraft está cheio de enguias',
    'pt-br': 'Meu aerodeslizador está cheio de enguias',
    'pt-br-alt': 'Meu aerobarco está cheio de enguias',
    'ro': 'Aeroglisorul meu e plin de țipari',
    'ru': 'Моё су́дно на возду́шной поду́шке по́лно угре́й',
    'sm': "Ua tumu la'u ato fagota i pusi",
    'sc': 'Su hovercraft est penu de ambidas',
    'sr': 'Мој ховеркрафт је пун јегуља',
    'sn': 'Hovercraft yangu yakazara nemikunga',
    'si': 'මාගේ වායු පා යානයේ ආඳන් පිරී ඇත',
    'si-alt': 'මගේ වායු පා යානෙ ආඳො පරීලා',
    'sk': 'Moje vznášadlo je plné úhorov',
    'sl': 'Moje vozilo na zračni blazini je polno jegulj',
    'so': 'Huufarkarafkayga waxaa ka buuxa eels',
    'es': 'Mi aerodeslizador está lleno de anguilas',
    'su': 'Kapal ngalayang abdi pinuh ku belut',
    'sw': 'Gari langu linaloangama limejaa na mikunga',
    'sv': 'Min svävare är full med ålar',
    'tl': 'Puno ng palos ang aking hoberkrap',
    'ta-alt': 'என் மிதவை நிறைய விலாங்கு மீன்கள்',
    'ta': 'என் விமானம் நிறைய விலாங்கு மீன்கள்',
    'te': 'నా విమానము అంతా మలుగు చేపలతో నిండిపోయింది',
    'th': 'โฮเวอร์คราฟท์ของผมเต็มไปด้วยปลาไหล',
    'th-alt': 'โฮเวอร์คราฟท์ของฉันเต็มไปด้วยปลาไหล',
    'bo': 'ངའི་རླུང་འདེགས་གྲུ་གཟིངས་ནི་ཉ་མན་ལི་ཡིས་གང་གི་འདུག །',
    'ti': 'ናተይ ሆበርክራፍት ሙሉእ እዩ ብዓሣ',
    'to': "Hoku vakapuna tētē 'i 'olunga fono 'i he toke.",
    'tn': 'Mokoro wa me o tletse ditlhapi',
    'tr-alt': 'Hoverkraftım yılan balığı dolu',
    'tr': 'Hoverkraftımın içi yılan balığı dolu',
    'tvl': 'Taku hovercraft kō fonu i pusi.',
    'uk': 'Моє судно на повітряній подушці наповнене вуграми',
    'uz': "Mening havo yostiqli kemam ilonbalig'i bilan to'lgan",
    'vi-alt': 'Tàu cánh ngầm của tôi đầy lươn',
    'vi': 'Nói một thứ tiếng thì không bao giờ đủ',
    'cy': 'Mae fy hofrenfad yn llawn llyswennod',
    'wo': "Wolof   Sama aéroglisseur fèèss neu ak anguilles'",
    'yi': 'מײַן שוועבשיף איז פֿול מיט ווענגערס',
    'yi-alt': 'מײַן פּראָם איז פֿול מיט ווענגערס',
    'xh': 'Inqwelo etshitshiliza phezu kwamanzi izele ziipalanga',
    'yo': 'Ọkọ afategun-sare mi kun fun ẹja arọ',
    'zu': 'Umkhumbi wami ugcwele ngenyoka zemanzini',
}


# -*- coding: utf-8 -*-
#
#    TypeAtlas Options Store
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#

"""Implementation of program options.

Use Options.getInstance() to get the options.
Use OptionsWidget.getInstance() to get the configuration dialog.
"""

from typeatlas.compat import QtCore, QtGui, QtWidgets, QtModelProxies
from typeatlas.compat import Qt, Slot, Signal, setResizeMode
from collections import namedtuple
from typeatlas.langutil import _, N_, H_, U_, textlang
from typeatlas.uitools import iconSizes, iconSize, generalWidth, generalHeight
from typeatlas.uitools import getIcon
from collections.abc import Callable
from collections import OrderedDict
from typeatlas.util import OrderedSet, Overwriter, generic_type
from typeatlas.util import MaybeLazy
from typeatlas import external, proginfo
import configparser
import shutil
import io
import os
import json
import binascii
from functools import partial

SequenceOf = generic_type('Sequence')
Union = generic_type('Union')
Optional = generic_type('Optional')
Literal = generic_type('Literal')
Any = generic_type('Any')

Choice = namedtuple('Choice', 'value label')
Choice.icon = None

ChoiceType = Union[Choice, Any]

DEFAULT = object()
DefaultType = Union[bool, Literal[DEFAULT]]


_checkboxMap = {Qt.Checked: True, Qt.Unchecked: False,
                Qt.PartiallyChecked: DEFAULT}


### FIXME FIXME FIXME: Default for pixels and stuff should be *NOT PRESENT*
###                    So that when we get bigger monitor, it grows?


class _Args(tuple):
    """A tuple of *args to call the type/class with.

    When returned by a type factory, it means calling the type
    itself with the given *args."""

    __slots__ = ()


def _parseSizeOrPoint(delim: str, factory: Callable, s: str) -> _Args:
    """Parse point or size, separated by a separator like 'x' or ','.
    Return the arguments for QSize or QPoint. You need to provide a
    number factory, like int or float."""

    parts = s.split('x')
    if len(parts) == 1:
        sz = factory(s)
        return _Args((sz, sz))

    if len(parts) == 2:
        return _Args(map(factory, parts))

    raise ValueError("Invalid size or point %r" % (s, ))


# Parsers for the given types
parsers = {
    QtCore.QSize: partial(_parseSizeOrPoint, 'x', int),
    QtCore.QSizeF: partial(_parseSizeOrPoint, 'x', float),
    QtCore.QPoint: partial(_parseSizeOrPoint, ',', int),
    QtCore.QPointF: partial(_parseSizeOrPoint, ',', float),
    bytes: binascii.a2b_base64,
    object: json.loads,
    dict: json.loads,
    list: json.loads,
}


# Dumpers for the given types
dumpers = {
    QtCore.QSize: lambda v: '%sx%s' % (v.width(), v.height()),
    QtCore.QSizeF: lambda v: '%sx%s' % (v.width(), v.height()),
    QtCore.QPoint: lambda v: '%s, %s' % (v.x(), v.y()),
    QtCore.QPointF: lambda v: '%s, %s' % (v.x(), v.y()),
    bytes: lambda v: binascii.b2a_base64(v, newline=False).decode('ascii'),
    object: partial(json.dumps, ensure_ascii=False),
    dict: partial(json.dumps, ensure_ascii=False),
    list: partial(json.dumps, ensure_ascii=False),
}


class Option(object):

    """Config file option."""

    def __init__(self, name: str, section: str, label: str=None,
                       type: type=None,
                       choices: MaybeLazy[SequenceOf[ChoiceType]]=None,
                       default: MaybeLazy[Any]=None,
                       icon: str=None, hasDefault: bool=None,
                       unit: str=None, attribute: str=None,
                       signalAttribute: str=None,
                       fromString: Callable=None, parseString: Callable=None,
                       toString: Callable=None):

        if hasDefault is None:
            #hasDefault = True if choices oeelse False
            hasDefault = True
        self.name = name
        self.section = section
        self.type = type
        self.choices = choices
        self.label = label
        self.icon = icon
        self.default = default
        self.hasDefault = hasDefault

        if parseString is None:
            parseString = parsers.get(type)
        if toString is None:
            toString = dumpers.get(type)

        self.parseString = parseString

        if toString is not None:
            self.toString = toString
        if fromString is not None:
            self.fromString = fromString

        if attribute is None:
            attribute = self.makeAttributeName(name)
        if signalAttribute is None:
            signalAttribute = attribute + 'Changed'
        self.attribute = attribute
        self.signalAttribute = signalAttribute

    def fromString(self, value: str) -> Any:
        """Parse the option value from string, returning a value of
        whose type self.type. This can be overriden with fromString=
        in the constructor. The parseString= can override just the
        parsing, without the construction."""

        if self.parseString is not None:
            value = self.parseString(value)
        if self.type is not None:
            if isinstance(value, _Args):
                value = self.type(*value)
            else:
                value = self.type(value)
        return value

    def toString(self, value: Any) -> str:
        """Convert the value to string. The default implementation just
        calls str() on the value, you can override with toString=
        in the constructor."""
        return str(value)

    def getChoices(self) -> SequenceOf[ChoiceType]:
        """Get the available choices."""
        choices = self.choices
        if choices is FONT_SIZES:
            return QtGui.QFontDatabase.standardSizes()
        if callable(choices):
            choices = choices()
        return choices

    def property(option: 'Option', optionCls: type=None) -> property:
        """Create a property that gets/sets that option through the Options
        object."""
        def getter(self):
            try:
                return self.getValue(option.name, option.section)
            except KeyError:
                raise AttributeError

        def setter(self, value):
            self.setValue(option.name, option.section, value)

        getter.__name__ = option.attribute
        setter.__name__ = option.attribute

        return property(getter, setter)

    def signal(self) -> Signal:
        """Create a signal descriptor that listens for value changes of this option,
        to be placed in the the Options class."""
        if self.type is bytes:
            return Signal(QtCore.QByteArray)
        return Signal(self.type or object)

    @classmethod
    def makeAttributeName(self, value: str) -> str:
        """Return a suitable attribute name for a given option."""
        parts = value.split('-')
        if len(parts) < 2:
            return value
        return parts[0] + ''.join(part.capitalize() for part in parts[1:])

    def _makeChoice(self, choice: ChoiceType,
                          default: MaybeLazy[DefaultType]=False,
                          realValue: Any=None) -> Choice:
        """Return a Choice() object for the given choice."""

        if isinstance(choice, Choice):
            return choice
        if choice is DEFAULT:
            if realValue is not None:
                return Choice(choice, '%s (%s)'
                                % (N_('Default'), self.toString(realValue)))
            return Choice(choice, N_('Default'))
        if choice is None:
            return Choice(choice, N_('Default') if default else N_('Unset'))
        return Choice(choice, self.toString(choice))

    def getCheckbox(self, options: 'OptionsBase',
                          parent: QtWidgets.QWidget=None) -> QtWidgets.QCheckBox:
        """Get a checkbox for this option."""
        checkbox = QtWidgets.QCheckBox(parent)

        checkbox.setText(_(self.label))
        if self.icon:
            checkbox.setIcon(getIcon(self.icon))

        value = options.getRawValue(self.name, self.section)

        default = self.default
        if callable(default):
            default = default()

        if self.hasDefault:
            checkbox.setTristate(True)
            if value is DEFAULT:
                checkbox.setCheckState(Qt.PartiallyChecked)

        if value is not DEFAULT:
            if value:
                checkbox.setCheckState(Qt.Checked)
            else:
                checkbox.setCheckState(Qt.Unchecked)

        @checkbox.stateChanged.connect
        def callback(state):
            options.setDraftValue(self.name, self.section,
                                  _checkboxMap.get(state))
        return checkbox

    def getComboBox(self, options: 'OptionsBase',
                          parent: QtWidgets.QWidget=None) -> QtWidgets.QComboBox:
        """Return a combo box for this option."""
        combo = QtWidgets.QComboBox(parent)
        if self.type is None:
            combo.setEditable(False)
        else:
            combo.setEditable(True)

        values = []

        currentValue = options.getRawValue(self.name, self.section)

        if (currentValue is not None and currentValue is not DEFAULT and
            combo.isEditable()):
                combo.setEditText(self.toString(currentValue))

        if self.hasDefault:

            default = self.default
            if callable(default):
                default = default()

            defaultChoice = self._makeChoice(DEFAULT, default=True,
                                             realValue=default)
            values.append(DEFAULT)

            if combo.isEditable():
                combo.lineEdit().setPlaceholderText(_(defaultChoice.label))

            if defaultChoice.icon:
                combo.addItem(getIcon(defaultChoice.icon),
                              _(defaultChoice.label))
            else:
                combo.addItem(_(defaultChoice.label))

            combo.setEditText('')

        for i, choice in enumerate(self.getChoices() or (), 1):
            choice = self._makeChoice(choice, default=False)
            values.append(choice.value)
            if choice.icon:
                combo.addItem(getIcon(choice.icon), _(choice.label))
            else:
                combo.addItem(_(choice.label))

            if currentValue is not DEFAULT and currentValue == choice.value:
                combo.setCurrentIndex(i)

        @combo.currentIndexChanged.connect
        def indexCallback(index: int):
            if index == -1:
                return
            value = values[index]
            options.setDraftValue(self.name, self.section, value)
            if not combo.isEditable():
                return
            if value is DEFAULT or value is None:
                combo.setEditText('')
            else:
                combo.setEditText(self.toString(value))

        def editCallback():
            value = combo.currentText()
            if not value:
                options.setDraftValue(self.name, self.section, DEFAULT)
                return
            try:
                value = self.fromString(value)
            except (ValueError, TypeError) as exc:
                ## TODO: Changing the index calls this?
                options.setDraftError(self.name, self.section, exc)

            else:
                # TODO: Should be automatic
                # TODO: Better handling of invalid values?
                ##options.clearDraftError(self.name, self.section)
                options.setDraftValue(self.name, self.section, value)

        if combo.isEditable():
            combo.lineEdit().editingFinished.connect(editCallback)

        label = QtWidgets.QLabel()
        label.setTextFormat(Qt.PlainText)
        label.setText(_(self.label) + ':')

        box = QtWidgets.QWidget()
        box.setLayout(QtWidgets.QHBoxLayout())
        box.layout().addWidget(label)
        box.layout().addWidget(combo)

        return box

    def getWidget(self, options: 'OptionsBase',
                        parent: QtWidgets.QWidget=None) -> QtWidgets.QWidget:
        """Return a widget to edit this option."""
        if self.choices:
            return self.getComboBox(options, parent)
        elif issubclass(self.type, bool):
            return self.getCheckbox(options, parent)
        else:
            # TODO: LineEdit?
            return self.getComboBox(options, parent)


FONT_SIZES = lambda: QtGui.QFontDatabase.standardSizes()

options = [

    Option('type-atlas-size', 'window', type=QtCore.QSize,
           default=lambda: QtCore.QSize(generalWidth(refSize=1200),
                                        generalHeight(refSize=800))),
    Option('font-sampler-size', 'window', type=QtCore.QSize,
           default=lambda: QtCore.QSize(generalWidth(refSize=1200),
                                        generalHeight(refSize=800))),
    Option('font-duel-size', 'window', type=QtCore.QSize,
           default=lambda: QtCore.QSize(generalWidth(refSize=800),
                                        generalHeight(refSize=600))),

    Option('glyph-atlas-size', 'window', type=QtCore.QSize,
           default=lambda: QtCore.QSize(generalWidth(refSize=800),
                                        generalHeight(refSize=600))),

    Option('type-atlas-state', 'window', type=bytes, default=b''),
    Option('type-atlas-arrangement', 'window', type=str, default='simple'),
    Option('type-atlas-splitter-state', 'window', type=bytes, default=b''),
    Option('type-atlas-dock-inner-state', 'window', type=dict, default=lambda: {}),
    Option('type-atlas-pane-sizes', 'window', type=dict, default=lambda: {}),

    Option('toolbar-icon-size', 'display',
           N_('Toolbar icon size'),
           type=QtCore.QSize, choices=iconSizes, unit=N_('pixels'),
           default=lambda: iconSize(refSize=16)),
    Option('info-icon-size', 'display',
           N_('Information document icon size'),
           type=QtCore.QSize, choices=iconSizes, unit=N_('pixels'),
           default=lambda: iconSize(refSize=16)),
    Option('info-flag-size', 'display',
           N_('Information document flag size'),
           type=QtCore.QSize, choices=iconSizes, unit=N_('pixels'),
           default=lambda: iconSize(refSize=16)),
    Option('tooltip-icon-size', 'display',
           N_('Tooltip preview icon size'),
           type=QtCore.QSize, choices=iconSizes, unit=N_('pixels'),
           default=lambda: iconSize(refSize=32)),

    Option('list-font-size', 'display',
           N_('List font size'),
           type=int, choices=FONT_SIZES, unit=N_('pt'), default=16),
    Option('sample-font-size', 'display',
           N_('Default sample font size'),
           type=int, choices=FONT_SIZES, unit=N_('pt'), default=32),
    Option('glyph-stack-font-size', 'display',
           N_('Font size of stack comparison'),
           type=int, choices=FONT_SIZES, unit=N_('pt'), default=96),
    Option('preview-font-size', 'display',
           N_('Default preview font size'),
           type=int, choices=FONT_SIZES, unit=N_('pt'), default=16),
    Option('grid-font-size', 'display',
           N_('Default font grid font size'),
           type=int, choices=FONT_SIZES, unit=N_('pt'), default=32),
    Option('char-box-font-size', 'display',
           N_('Font grid preview character size'),
           type=int, choices=FONT_SIZES, unit=N_('pt'), default=48),

    Option('standard-checkboxes', 'display',
           N_("Use standard OS style for all checkboxes"),
           type=bool, default=False),

    Option('qt-version', 'runtime', N_('Preferred Qt version'),
            choices=[
                ## Not really supported, some features (e.g. tags, categories)
                ## require 5 (they can be made usable-ish with 4).
                #Choice('4', U_('Qt 4')),
                Choice('5', U_('Qt 5')),
            ]),
    Option('qt-bindings', 'runtime', N_('Preferred Qt bindings'),
            choices=[Choice('pyqt', U_('PyQt / SIP (Riverbank)')),
                     Choice('pyside', U_('PySide / Shiboken (Qt)'))]),
    Option('qt-style', 'runtime', N_('Qt style'),
           choices=lambda: QtWidgets.QStyleFactory.keys(),
           type=str),

    Option('sample-selection', 'samples', N_('Font sample selection'),
           choices=[Choice('some', N_('Display highlights')),
                    Choice('most', N_('Display all available samples'))]),
    Option('show-sample-name', 'samples', N_('Display the sample name'),
           type=bool, default=True),

    Option('max-sample-fonts', 'samples',
           N_('Maximum fonts in preview'),
           type=int, default=100),
    Option('max-font-grid-fonts', 'samples',
           N_('Maximum fonts in character map'),
           type=int, default=15),

    Option('layered-font-renderer', 'routines',
           N_('Renderer for color layer fonts (CPAL/COLR)'),
           choices=[Choice('os', N_('Operating system')),
                    Choice('internal', N_('Internal renderer'))]),
    Option('svg-font-renderer', 'routines',
           N_('Renderer for SVG color fonts'),
           choices=[Choice('os', N_('Operating system')),
                    Choice('internal', N_('Internal renderer'))]),


]


class OptionsWidget(QtWidgets.QDialog):

    """Edit options dialog window / widget."""

    def __init__(self, options: 'OptionsBase'=None,
                       parent: QtWidgets.QWidget=None):
        super().__init__(parent)

        if options is None:
            options = Options.getInstance()

        self.options = options

        self.setWindowTitle(_('TypeAtlas options'))

        self.general = QtWidgets.QWidget()
        self.generalLayout = layout = QtWidgets.QVBoxLayout()
        self.general.setLayout(self.generalLayout)

        iconLayout = QtWidgets.QVBoxLayout()
        iconBox = QtWidgets.QGroupBox(_('Icon sizes'))
        iconBox.setLayout(iconLayout)
        layout.addWidget(iconBox)

        fontLayout = QtWidgets.QVBoxLayout()
        fontBox = QtWidgets.QGroupBox(_('Font sizes'))
        fontBox.setLayout(fontLayout)
        layout.addWidget(fontBox)

        iconLayout.addWidget(options.getWidget('toolbar-icon-size'))
        iconLayout.addWidget(options.getWidget('info-icon-size'))
        iconLayout.addWidget(options.getWidget('info-flag-size'))
        iconLayout.addWidget(options.getWidget('tooltip-icon-size'))
        fontLayout.addWidget(options.getWidget('list-font-size'))
        fontLayout.addWidget(options.getWidget('sample-font-size'))
        fontLayout.addWidget(options.getWidget('preview-font-size'))
        fontLayout.addWidget(options.getWidget('grid-font-size'))
        fontLayout.addWidget(options.getWidget('char-box-font-size'))
        layout.addWidget(options.getWidget('standard-checkboxes'))
        layout.addWidget(options.getWidget('qt-version'))
        layout.addWidget(options.getWidget('qt-bindings'))
        layout.addWidget(options.getWidget('qt-style'))

        self.behaviour = QtWidgets.QWidget()
        self.behaviourLayout = layout = QtWidgets.QVBoxLayout()
        self.behaviour.setLayout(self.behaviourLayout)

        sampleLayout = QtWidgets.QVBoxLayout()
        sampleBox = QtWidgets.QGroupBox(_('Samples'))
        sampleBox.setLayout(sampleLayout)
        layout.addWidget(sampleBox)

        sampleLayout.addWidget(options.getWidget('sample-selection'))
        sampleLayout.addWidget(options.getWidget('show-sample-name'))
        sampleLayout.addWidget(options.getWidget('max-sample-fonts'))
        sampleLayout.addWidget(options.getWidget('max-font-grid-fonts'))

        colorFontLayout = QtWidgets.QVBoxLayout()
        colorFontBox = QtWidgets.QGroupBox(_('Color fonts'))
        colorFontBox.setLayout(colorFontLayout)
        layout.addWidget(colorFontBox)

        colorFontLayout.addWidget(options.getWidget('layered-font-renderer'))
        colorFontLayout.addWidget(options.getWidget('svg-font-renderer'))

        checkbox = QtWidgets.QCheckBox(_('Enable 3D acceleration'))
        checkbox.setDisabled(True)
        colorFontLayout.addWidget(checkbox)

        self._executableForEdit = {}

        self.externalBox = QtWidgets.QWidget()
        self.externalLayout = QtWidgets.QGridLayout()
        self.externalBox.setLayout(self.externalLayout)

        for i, executable in enumerate(external.commands_by_executable):
            defaultPath = shutil.which(executable)
            label = QtWidgets.QLabel()
            label.setTextFormat(Qt.PlainText)
            label.setText(executable)

            edit = QtWidgets.QLineEdit()
            if defaultPath:
                edit.setPlaceholderText(defaultPath)
            else:
                edit.setPlaceholderText(_('Select executable...'))

            self._executableForEdit[edit] = executable

            if hasattr(edit, 'setClearButtonEnabled'):
                edit.setClearButtonEnabled(True)

            execPath = options.executablePaths.get(executable)
            if execPath:
                edit.setText(execPath)

            edit.editingFinished.connect(self.pathEditEdited)


            button = QtWidgets.QPushButton(_("Browse..."))
            button.clicked.connect(partial(self.browse, edit, defaultPath))

            self.externalLayout.addWidget(label, i, 0, 1, 1)
            self.externalLayout.addWidget(edit, i, 1, 1, 1)
            self.externalLayout.addWidget(button, i, 2, 1, 1)

        self.tabs = QtWidgets.QTabWidget()
        self.tabs.addTab(self.general, _('Interface'))
        self.tabs.addTab(self.behaviour, _('Behaviour'))
        self.tabs.addTab(self.externalBox, _('Dependencies'))

        self.buttonBox = QtWidgets.QDialogButtonBox()

        save = QtWidgets.QPushButton(getIcon('document-save'), _('Save'))
        cancel = QtWidgets.QPushButton(getIcon('dialog-cancel'), _('Cancel'))

        self.buttonBox.addButton(save, self.buttonBox.AcceptRole)
        self.buttonBox.addButton(cancel, self.buttonBox.RejectRole)

        save.clicked.connect(self.accept)
        cancel.clicked.connect(self.reject)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().addWidget(self.tabs)
        self.layout().addWidget(self.buttonBox)

        self.accepted.connect(options.save)
        self.rejected.connect(options.rollback)

        self.completed = False

    closed = Signal()

    @Slot()
    def pathEditEdited(self):
        """A path of an external executable was edited. This
        needs to be called from a QLineEdit signal."""

        edit = self.sender()

        executable = self._executableForEdit[edit]
        path = edit.text()

        if not path:
            self.options.executablePaths.pop(executable, None)
        else:
            self.options.executablePaths[executable] = path

    def browse(self, edit: QtWidgets.QLineEdit, defaultPath: str):
        """Browse for an executable's path."""

        dialog = QtWidgets.QFileDialog()
        if defaultPath:
            directory, filename = os.path.split(defaultPath)
            dialog.setDirectory(directory)
            dialog.selectFile(filename)

        dialog.setFilter(QtCore.QDir.Executable | QtCore.QDir.Files)
        if (dialog.exec_()):
            filename = dialog.selectedFiles()[0]
            edit.setText(filename)

    def closeEvent(self, event: QtCore.QEvent):
        r = super().closeEvent(event )
        if event.isAccepted():
            self.closed.emit()
        return r


class OptionsBase(QtCore.QObject):

    """A base class for options that has the implementations, without the
    option list. The options are in the options attribute.

    Use Options.getInstance().
    """

    options = OrderedDict()

    instance = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        values = {option.name: DEFAULT for option in self.options.values()}

        self.temporaryValues = dict(values)
        self.permanentValues = dict(self.temporaryValues)

        self.parser = None
        self.changed = False
        self.pending = False

    @classmethod
    def getInstance(cls) -> 'OptionsBase':
        """Return the singleton instance of the options, or create
        one if there is not one created yet."""
        if cls.instance is None:
            cls.instance = cls()
        return cls.instance

    def fileName(self, create: bool=False) -> Optional[str]:
        """Return the file name of the options. Implement in a subclass."""
        if create:
            raise NotImplementedError
        return None

    @Slot()
    def save(self, file: io.TextIOBase=None):
        """Save the options. If a file is specified, they are saved to it
        instead of the default one."""
        if file is None:
            with Overwriter(self.fileName(create=True), 'w',
                            encoding='utf8') as file:
                self.save(file)
            return

        if self.parser is None:
            parser = configparser.RawConfigParser()
        else:
            parser = self.parser

        self.saveIntoParser(parser)

        parser.write(file)

    @Slot()
    def load(self, file: io.TextIOBase=None):
        """Load the options. If a file is specified, they are loaded from it
        instead from the default one."""
        if file is None:
            filename = self.fileName()
            if filename is None:
                return
            with open(filename, 'r', encoding='utf8') as file:
                self.load(file)
            return

        parser = configparser.RawConfigParser()

        parser.read_file(file)


        self.loadFromParser(parser)

        self.parser = parser

    def makeSectionName(self, section: str) -> str:
        """Format the section name for the config file."""
        return section.capitalize()

    def makeOptionName(self, option: str) -> str:
        """Format the option name for the config file."""
        return option.replace('-', '_')

    def saveIntoParser(self, parser: configparser.RawConfigParser):
        """Save the options into the given config parser object."""
        self.commit()
        sections = OrderedSet(opt.section for opt in self.options.values())

        for section in sections:
            section = self.makeSectionName(section)
            if not parser.has_section(section):
                parser.add_section(section)

        for option in self.options.values():
            section = self.makeSectionName(option.section)
            optionName = self.makeOptionName(option.name)
            if self.permanentValues[option.name] is DEFAULT:
                if parser.has_option(section, optionName):
                    parser.remove_option(section, optionName)
            else:
                parser.set(section, optionName,
                           option.toString(self.permanentValues[option.name]))

        self.changed = False
        self.pending = False

    def loadFromParser(self, parser: configparser.RawConfigParser):
        """Load the options from the given config parser."""
        for option in self.options.values():
            section = self.makeSectionName(option.section)
            optionName = self.makeOptionName(option.name)
            if not parser.has_option(section, optionName):
                continue
            value = parser.get(section, optionName)
            value = option.fromString(value)
            self.permanentValues[option.name] = value
            self.temporaryValues[option.name] = value

        self.changed = False
        self.pending = False

    @classmethod
    def makeClass(cls, name: str, options: SequenceOf[Option]) -> type:
        """Create a class for options using a given list of such."""
        namespace = {}
        namespace['options'] = OrderedDict()
        for option in options:
            namespace['options'][option.name] = option
            namespace[option.attribute] = option.property()
            namespace[option.signalAttribute] = option.signal()

        return type(name, (cls, ), namespace)

    @Slot()
    def commit(self):
        """Confirm any temporary changes. This happens when e.g. the user
        clicks OK in the options dialog. It is automatically triggered
        when one saves them.

        Editing options usually uses setDraftValue() to set a temporary
        value which is confirmed on save, or rolled back on Cancel.
        """
        self.changed = True
        self.permanentValues = dict(self.temporaryValues)
        self.pending = False

    @Slot()
    def rollback(self):
        """Reset any temporary changes back to the values before the change.

        Editing options usually uses setDraftValue() to set a temporary
        value which is confirmed on save, or rolled back on Cancel.

        This triggers signals for the rolled back options.
        """
        sentinel = object()
        for option in self.options.values():
            value = self.temporaryValues[option.name]
            oldValue = self.permanentValues[option.name]
            if value != oldValue:
                self.setDraftValue(option.name, option.section, oldValue)
        self.pending = False

    def setDraftValue(self, option: str, section: str, value: object):
        """Set the draft value for the given option in the given section.

        This triggers signals for the edited options.

        Editing options usually uses setDraftValue() to set a temporary
        value which is confirmed on save, or rolled back on Cancel.
        """
        oldValue = self.temporaryValues[option]

        if value == oldValue:
            return

        self.pending = True

        self.temporaryValues[option] = value

        if value is DEFAULT:
            value = self.options[option].default
            if callable(value):
                value = value()

        getattr(self, self.options[option].signalAttribute).emit(value)

    def setValue(self, option: str, section: str, value: object):
        """Set the value for a given option in a given section.
        This is automatcailly invoked when setting the property.

        This triggers signals for the edited options."""
        self.changed = True
        self.permanentValues[option] = value
        self.setDraftValue(option, section, value)

    def getWidget(self, name: str, *args, **kwargs) -> QtWidgets.QWidget:
        """Get the widget for the given option. All arguments are
        passed to the options's getWidget()."""
        return self.options[name].getWidget(self, *args, **kwargs)

    def getComboBox(self, name: str, *args, **kwargs) -> QtWidgets.QComboBox:
        """Get the combo box for the given option. All arguments are
        passed to the options's getComboBox()."""
        return self.options[name].getComboBox(self, *args, **kwargs)

    def getCheckbox(self, name: str, *args, **kwargs) -> QtWidgets.QCheckBox:
        """Get the check box for the given option. All arguments are
        passed to the options's getCheckbox()."""
        return self.options[name].getCheckbox(self, *args, **kwargs)

    def getValue(self, option: str, section: str) -> object:
        """Get the value of a given option. This always returns the temporary
        value. Or the default, if no value is set."""
        value = self.temporaryValues[option]
        if value is DEFAULT:
            default = self.options[option].default
            if callable(default):
                default = default()
            return default
        return value

    def getRawValue(self, option: str, section: str) -> object:
        """Get the raw value of a given option, without getting the default.
        This always returns the temporary value, and raises a KeyError if one
        is not set."""
        return self.temporaryValues[option]


class Options(OptionsBase.makeClass('Options', options)):

    """The options of typeatlas. Use Options.getInstance() to get them."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.executablePaths = {}
        self.qtStyleChanged.connect(self._qtStyleChanged)

    def fileName(self, create: bool=False) -> str:
        if create:
            if not os.path.exists(proginfo.CONFIG_DIR):
                os.makedirs(proginfo.CONFIG_DIR)
        filename = os.path.join(proginfo.CONFIG_DIR, proginfo.MAIN_CONFIG)
        if not os.path.exists(filename) and not create:
            filename = None
        return filename

    def loadFromParser(self, parser: configparser.RawConfigParser):
        super().loadFromParser(parser)

        section = 'ExternalTools'

        if not parser.has_section(section):
            return

        for option in parser.options(section):
            self.executablePaths[option] = parser.get(section, option)

    def saveIntoParser(self, parser: configparser.RawConfigParser):
        super().saveIntoParser(parser)

        section = 'ExternalTools'

        if not parser.has_section(section):
            parser.add_section(section)

        for option in parser.options(section):
            if option not in self.executablePaths:
                parser.remove_option(section, option)

        for executable, path in self.executablePaths.items():
            parser.set(section, executable, path)

    @Slot(str)
    def _qtStyleChanged(self, style: str):
        """Called when the Qt style has changed in the options. This
        triggers a change in the GUI."""
        QtWidgets.QApplication.setStyle(style)





# -*- coding: utf-8 -*-
#
#    TypeAtlas GUI Common Functions/Widgets
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#

"""Common GUI widgets used across TypeAtlas."""

from typeatlas.compat import QtCore, QtGui, QtWidgets, QtModelProxies
from typeatlas.compat import Qt, Slot, Signal, setResizeMode
from typeatlas.uitools import Toolbox, iconSize, getIcon, DragDropWidget
from typeatlas.uitools import getIconHtml, getImage
from typeatlas.uitools import generalWidth, generalHeight
from typeatlas.uitools import TYPING_DELAY
from typeatlas.options import Options
from typeatlas.charinfo import CharacterDatabase, all_character_codes
from typeatlas.langutil import LanguageDatabase
from typeatlas.langutil import _, N_, H_, textlang
from typeatlas import blockmath, charinfo, osinfo, annotations
from functools import partial
from itertools import groupby
from operator import attrgetter
from collections import OrderedDict, Counter
from collections.abc import Callable, Sequence, Iterator
from typeatlas.util import OrderedSet, format_size, generic_type
import typeatlas
import os


Union = generic_type('Union')
Optional = generic_type('Optional')
IterableOf = generic_type('Iterable')
SequenceOf = generic_type('Sequence')


class CustomComboBox(QtWidgets.QComboBox):

    """A combo box with extra features, such as it can be made narrow,
    as required by the unicode block and script combo boxes. It provides
    signals that are emitted before and after showing the popup, which
    is also needed for the aforementioned combo boxes.."""

    popupAboutToShown = Signal()
    popupShown = Signal()

    narrow = False

    def setNarrow(self, narrow: bool=True,
                        sizeTextHint: str='Six words long'):
        """Make the combo box narrow, using the text hint to determine the
        size. On show, the size may be reset to actual contents."""
        self.narrow = narrow
        if narrow:
            metrics = QtGui.QFontMetrics(QtGui.QFont())

            self.setSizeAdjustPolicy(self.AdjustToMinimumContentsLength)
            # That's actually ignored?
            self.setMinimumSize(QtCore.QSize(metrics.width(sizeTextHint),
                                             metrics.height()))
        else:
            self.setSizeAdjustPolicy(self.AdjustToContentsOnFirstShow)
            self.setMinimumSize(QtCore.QSize(0, 0))

    def showPopup(self):
        """Show the popup."""
        if self.narrow:
            self.view().setMinimumWidth(self.view().sizeHintForColumn(0))
        self.popupAboutToShown.emit()
        super().showPopup()
        self.popupShown.emit()



class FontRenderingChoice(Toolbox):

    """Menu items for setting the font rendering. Since this uses the options,
    some modules incorrectly use this class to carry the options instance
    around."""

    def __init__(self, options: Options=None):
        if options is None:
            options = Options.getInstance()
        self.antialiasStyle = QtGui.QFont.PreferDefault
        self.hintingPreference = QtGui.QFont.PreferDefaultHinting

        self.listSize = options.listFontSize
        self.sampleSize = options.sampleFontSize
        self.glyphStackSize = options.glyphStackFontSize
        self.previewSize = options.previewFontSize
        self.gridSize = options.gridFontSize
        self.charBoxSize = options.charBoxFontSize

        self.options = options
        self.antialiasingItems = [
            (QtGui.QFont.PreferDefault, _('Default antialiasing')),
            (QtGui.QFont.NoAntialias, _('No antialiasing')),
        ]
        if hasattr(QtGui.QFont, 'NoSubpixelAntialias'):
            self.antialiasingItems.append(
                (QtGui.QFont.NoSubpixelAntialias, _('No subpixel rendering')))
        self.antialiasingItems.append(
            (QtGui.QFont.PreferAntialias, _('Antialiasing')))

        self.hintingItems = [
            (QtGui.QFont.PreferDefaultHinting, _('Default hinting')),
            (QtGui.QFont.PreferNoHinting, _('No hinting')),
            (QtGui.QFont.PreferVerticalHinting, _('Vertical hinting')),
            (QtGui.QFont.PreferFullHinting, _('Full hinting')),
        ]

        #self.antialiasStyle = QtGui.QFont.NoAntialias

        self._antialiasingActions = {}
        self._hintingActions = {}

        super(FontRenderingChoice, self).__init__()

        options.listFontSizeChanged.connect(self.setListSize)
        options.sampleFontSizeChanged.connect(self.setSampleSize)
        options.glyphStackFontSizeChanged.connect(self.setGlyphStackSize)
        options.previewFontSizeChanged.connect(self.setPreviewSize)
        options.gridFontSizeChanged.connect(self.setGridSize)
        options.charBoxFontSizeChanged.connect(self.setCharBoxSize)

    def actionDefinitions(self) -> Iterator:
        with self.group(exclusive=True):
            for i, (value, label) in enumerate(self.antialiasingItems):
                yield self.action(partial(self._antialiasChanged, i), label,
                                  checkable=True, checked=not i)
        yield self.separator()

        with self.group(exclusive=True):
            for i, (value, label) in enumerate(self.hintingItems):
                yield self.action(partial(self._hintingChanged, i), label,
                                  checkable=True, checked=not i)

    changed = Signal()

    listChanged = Signal()
    listSizeChanged = Signal(int)

    previewChanged = Signal()
    previewSizeChanged = Signal(int)

    samplesChanged = Signal()
    sampleSizeChanged = Signal(int)

    glyphStackChanged = Signal()
    glyphStackSizeChanged = Signal(int)

    gridChanged = Signal()
    gridSizeChanged = Signal(int)

    charBoxChanged = Signal()
    charBoxSizeChanged = Signal(int)

    @Slot(int)
    def setListSize(self, size: int):
        """Set the font size in the font list."""
        self.listSize = size
        self.listSizeChanged.emit(size)
        self.listChanged.emit()

    @Slot(int)
    def setPreviewSize(self, size: int):
        """Set the font size in the sample preview pane."""
        self.previewSize = size
        self.previewSizeChanged.emit(size)
        self.previewChanged.emit()

    @Slot(int)
    def setSampleSize(self, size: int):
        """Set the font size in sample font list"""
        self.sampleSize = size
        self.sampleSizeChanged.emit(size)
        self.samplesChanged.emit()

    @Slot(int)
    def setGlyphStackSize(self, size: int):
        """Set the font size in the glyph stack in font duel"""
        self.glyphStackSize = size
        self.glyphStackSizeChanged.emit(size)
        self.glyphStackChanged.emit()

    @Slot(int)
    def setGridSize(self, size: int):
        """Set the font size in the font grid"""
        self.gridSize = size
        self.gridSizeChanged.emit(size)
        self.gridChanged.emit()

    @Slot(int)
    def setCharBoxSize(self, size: int):
        """Set the font size in the character box below the font grid"""
        self.charBoxSize = size
        self.charBoxSizeChanged.emit(size)
        self.charBoxChanged.emit()

    def populateToolbarWithCombos(self, toolbar: QtWidgets.QToolBar):
        """Add antialiasing combo boxes to a toolbar. Why???"""
        antialiasWidget = QtWidgets.QComboBox()
        for item in self.antialiasingItems:
            antialiasWidget.addItem(item[1])
        antialiasWidget.currentIndexChanged.connect(self._antialiasChanged)

        hintingWidget = QtWidgets.QComboBox()
        for item in self.hintingItems:
            hintingWidget.addItem(item[1])
        hintingWidget.currentIndexChanged.connect(self._hintingChanged)
        toolbar.addWidget(antialiasWidget)
        toolbar.addWidget(hintingWidget)

    @Slot(int, bool)
    def _antialiasChanged(self, index: int, checked: bool=False):
        """The antialiasing has changed. Emit all signals"""
        self.antialiasStyle = self.antialiasingItems[index][0]
        self.changed.emit()
        self.previewChanged.emit()
        self.listChanged.emit()
        self.samplesChanged.emit()
        self.gridChanged.emit()

    @Slot(int, bool)
    def _hintingChanged(self, index: int, checked: bool=False):
        """The hinting has changed. Emit all signals"""
        self.hintingPreference = self.hintingItems[index][0]
        self.changed.emit()
        self.previewChanged.emit()
        self.listChanged.emit()
        self.samplesChanged.emit()
        self.gridChanged.emit()


class TrashDropWidget(DragDropWidget):

    """A trash icon drop widget. Dropping things here deletes them."""

    def __init__(self, *args, dropEnabled=True, stretch=False, **kwargs):
        super().__init__(*args, dropEnabled=dropEnabled, **kwargs)
        self.iconLabel = QtWidgets.QLabel()
        self.textLabel = QtWidgets.QLabel()

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().addWidget(self.iconLabel)
        self.layout().addWidget(self.textLabel)

        self.iconSize = iconSize(refSize=32)
        self.icon = getIcon('user-trash')

        self.iconLabel.setPixmap(self.icon.pixmap(self.iconSize))
        self.textLabel.setText(_('Trash'))

        if stretch:
            self.layout().addStretch(1)

    def setIcon(self, icon: QtGui.QIcon):
        """Set the icon of the trash can. QIcon.fromTheme('user-trash') is
        a good choice, and is the default."""
        self.icon = icon
        self.iconLabel.setPixmap(self.icon.pixmap(self.iconSize))

    def setIconSize(self, size: int):
        """Set the icon size. The default would be 32 pixels on
        a 96 DPI screen."""
        self.iconSize = size
        self.iconLabel.setPixmap(self.icon.pixmap(self.iconSize))

    def setLabelText(self, text: str):
        """Set the text label. Defaults to localized 'Trash'."""
        self.textLabel.setText(text)

    def supportedDropActions(self):
        if self.dropEnabled():
            return Qt.MoveAction
        else:
            return Qt.IgnoreAction


def _add_key_value_to_grid(grid: QtWidgets.QGridLayout,
                           name: str, value: str,
                           icon: Union[QtGui.QIcon, str]=None, *,
                           options: Options,
                           moreInfoSlot: Callable=None):
    """Add key and value row to a grid layout, with an optional icon.

    Options are required to get the font size. If a moreInfoSlot callable
    is provided, it will be called when an ellipsis button is pressed.

    Used by the AnnotationWidget.
    """

    i = grid.rowCount()

    nameLabel = QtWidgets.QLabel()
    valueLabel = QtWidgets.QLabel()

    nameLabel.setTextFormat(Qt.PlainText)
    valueLabel.setTextFormat(Qt.PlainText)

    nameLabel.setText(name + ':')
    valueLabel.setText(value)


    nameLabel.setAlignment(Qt.AlignVCenter | Qt.AlignRight)

    if icon is None or (not icon and icon == ''):
        grid.addWidget(nameLabel, i, 0, 1, 1)
        valuePos = (i, 1, 1, 2)
    else:
        if isinstance(icon, str):
            icon = getIcon(icon)
        iconLabel = QtWidgets.QLabel()
        iconLabel.setPixmap(icon.pixmap(options.infoIconSize))

        grid.addWidget(nameLabel, i, 0, 1, 1)
        grid.addWidget(iconLabel, i, 1, 1, 1)
        valuePos = (i, 2, 1, 1)

    if moreInfoSlot is None:
        grid.addWidget(valueLabel, *valuePos)

    else:
        more = QtWidgets.QPushButton('...')
        more.clicked.connect(moreInfoSlot)
        #more.setMinimumSize(QtCore.QSize(0, 0))
        more.setFlat(True)

        valueLayout = QtWidgets.QHBoxLayout()
        valueLayout.addWidget(valueLabel, 1)
        valueLayout.addWidget(more)

        grid.addLayout(valueLayout, *valuePos)


def _add_separator_to_grid(grid: QtWidgets.QGridLayout):
    """Add a separator row to the grid layout. Used by the
    AnnotationWidget."""

    sep = QtWidgets.QFrame()
    sep.setFrameShape(sep.HLine)
    grid.addWidget(sep, grid.rowCount(), 0, 1, 3)


class AnnotationWidget(QtWidgets.QWidget):

    """A widget displaying an annotation, usually pertaining
    to a font item."""

    def __init__(self, item: 'typeatlas.fontlist.FontLike',
                       info: annotations.Annotation,
                       isFileType: bool=False, *a, options: Options, **kw):
        super().__init__(*a, **kw)

        add = partial(_add_key_value_to_grid, options=options)
        add_separator = _add_separator_to_grid

        annotation = info.annotation
        grid = QtWidgets.QGridLayout()
        self.setLayout(grid)

        self.setWindowTitle(annotation.get_name(translate=_, decorated=True))
        self.setWindowIcon(getIcon(info.icon))

        trademarks = set()
        category = getattr(annotation, 'category', None)

        for relText, subannotation in annotation.relations(translate=_):

            if relText:
                add_separator(grid)

                note = QtWidgets.QLabel()
                note.setTextFormat(Qt.PlainText)
                note.setText(relText)
                grid.addWidget(note, grid.rowCount(), 0, 1, 3)

            nameTuples = subannotation.get_name_tuples(annotations.long,
                                                       translate=_,
                                                       decorated=True)

            for i, (header, text, nameOb) in enumerate(nameTuples):
                add(grid, header, text,
                    icon=subannotation.icon if not i else None)

            for noteText in subannotation.get_notes(translate=_):
                note = QtWidgets.QLabel()
                note.setTextFormat(Qt.PlainText)
                note.setText(noteText)
                grid.addWidget(note, grid.rowCount(), 0, 1, 3, Qt.AlignCenter)

            trademarks.update(subannotation.get_trademarks(translate=_))

        if category is not None:
            categories = annotation.categories()
        else:
            categories = ()

        for category in categories:
            add_separator(grid)
            add(grid, category.get_header(translate=_),
                        category.get_name(annotations.long, translate=_))

            if category.illustration:
                image = getImage(category.illustration)
                pixmap = image.pixmap(iconSize(refSize=256))
                illustration = QtWidgets.QLabel()
                illustration.setPixmap(pixmap)
                grid.addWidget(illustration, grid.rowCount(), 0, 1, 3,
                               Qt.AlignCenter)
            trademarks.update(category.get_trademarks(translate=_))

        if trademarks:
            add_separator(grid)
            for trademark in trademarks:
                note = QtWidgets.QLabel()
                note.setTextFormat(Qt.PlainText)
                note.setText(trademark)
                grid.addWidget(note, grid.rowCount(), 0, 1, 3, Qt.AlignCenter)

        grid.addWidget(QtWidgets.QWidget(), grid.rowCount(), 0, 1, 3)
        grid.setColumnStretch(grid.columnCount() - 1, 1)
        grid.setRowStretch(grid.rowCount() - 1, 1)


HIDE_FIRST_TAB = '''
            QTabBar::tab:first {
                max-width: 0px;
            }
        '''


class _TabRow(QtWidgets.QTabBar):

    """A tab row of a multi-tab widget. The first tab of this is a
    fake hidden tab widget."""

    def minimumTabSizeHint(self, index):
        if index == 0:
            res = QtCore.QSize(0, 0)
        else:
            res = super().minimumTabSizeHint(index)
        return res

    def tabSizeHint(self, index):
        if index == 0:
            res = QtCore.QSize(0, 0)
        else:
            res = super().tabSizeHint(index)
        return res

    _wheelMoved = Signal(QtGui.QWheelEvent)
    _mouseLeft = Signal(QtCore.QEvent)

    def wheelEvent(self, event):
        self._wheelMoved.emit(event)

    def leaveEvent(self, event):
        self._mouseLeft.emit(event)


class MultiRowTabWidget(QtWidgets.QWidget):

    """A multi-row tab widget.

    Despite Qt's opposition to such, the use of multi-row tab widgets by
    Microsoft (e.g. as seen in Microsoft Office 4.3) is nice. More
    importantly and less subjectively, it is less horrible than *hiding*
    the extra tabs and having to scroll to get to them, which totally
    hides the numbers of tabs available.

    Unfortunately, this widget tends to be ugly-ish with most themes, so
    we are not using it. Additionally, the rows need to be layed out manually.
    Perhaps it's easy to write a wrapper that does it automatically, depending
    on size.
    """

    currentChanged = Signal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._wheelTurn = 0

        self.currentRow = 0
        self.stack = QtWidgets.QStackedWidget()
        self.tabBars = []
        self.addTabBar()

    def addTabBar(self) -> int:
        """Add a new tab bar and return its row"""
        bar = _TabRow()
        bar.addTab('')
        bar.setExpanding(False)
        #bar.setShape(bar.TriangularNorth)
        bar.setStyleSheet(HIDE_FIRST_TAB)
        self.tabBars.append(bar)

        bar.currentChanged.connect(self._tabChanged)
        #bar.tabBarClicked.connect(self._tabClicked)
        bar._wheelMoved.connect(self._tabBarWheelMoved)
        bar._mouseLeft.connect(self._mouseLeftTabBar)

        self._relayout()
        return len(self.tabBars) - 1

    @Slot(int)
    def _tabClicked(self, tabIndex: int):
        """A tab was clicked. Reset wheel."""
        self._wheelTurn = 0

    @Slot(int)
    def _tabChanged(self, tabIndex: int):
        """The current tab changed. Reset wheel, change page,
        and optionally reorder bars."""
        self._wheelTurn = 0

        if not tabIndex:
            return

        col = tabIndex - 1
        bar = self.sender()
        row = self.tabBars.index(bar)
        if row != self.currentRow:
            oldBar = self.tabBars[self.currentRow]
            oldBar.setCurrentIndex(0)
            #oldBar.setDrawBase(False)
            self.currentRow = row
            #self.tabBars[row].setDrawBase(True)
            self._relayout()
        page = bar.tabData(tabIndex)

        self.stack.setCurrentWidget(page)
        self.currentChanged.emit(self.sindexOf(page))

    @Slot(QtCore.QEvent)
    def _mouseLeftTabBar(self, event: QtCore.QEvent):
        """Mouse left the tab bar. Reset wheel."""
        self._wheelTurn = 0

    @Slot(QtGui.QWheelEvent)
    def _tabBarWheelMoved(self, event: QtGui.QWheelEvent):
        """Wheel moved over tab bar. Switch tabs if needed."""
        try:
            delta = event.delta()

        except AttributeError:
            angleDelta = event.angleDelta()
            if angleDelta.isNull():
                return
            delta = max([angleDelta.y(), angleDelta.x()], key=abs)

        if not delta:
            return

        delta += self._wheelTurn
        self._wheelTurn = 0

        sign = +1 if delta > 0 else -1

        steps, remainder = divmod(abs(delta), 120)
        steps = int(steps)
        if steps:
            pageCount = self.count()
            if pageCount:
                self.setCurrentIndex((self.currentIndex() -
                                      sign * steps) % pageCount)

        self._wheelTurn = sign * remainder

    def currentIndex(self) -> int:
        """Get the index of the current selected tab page."""
        return self.stack.currentIndex()

    def count(self) -> int:
        """Get the total number of tab pages."""
        return self.stack.count()

    @Slot(int)
    def setCurrentIndex(self, index: int):
        """Set the index of the current selected tab page."""
        widget = self.widget(index)
        if not widget:
            return
        self.setCurrentWidget(widget)

    def widget(self, index: int) -> QtWidgets.QWidget:
        """Return the tab page at the given index."""
        return self.stack.widget(index)

    def indexOf(self, widget: QtWidgets.QWidget) -> int:
        """Get the index of the given widget."""
        return self.stack.indexOf(widget)

    @Slot(QtWidgets.QWidget)
    def setCurrentWidget(self, widget: QtWidgets.QWidget):
        """Set the current tab page by widget."""
        for bar in self.tabBars:
            for i in range(1, bar.count()):
                if bar.tabData(i) == widget:
                    bar.setCurrentIndex(i)
                    break

    def tabBar(self, row: int=0) -> _TabRow:
        """Return the tab bar for the given row, creating it
        if it does not exist yet.."""
        while row >= len(self.tabBars):
            self.addTabBar()
        return self.tabBars[row]

    def addTab(self, page: QtWidgets.QWidget, *args,
                     row: int=0, **kwargs) -> int:
        """Add a tab to a given row. The arguments are the
        same as for QTabWidget.addTab(), except for the row keyword
        argument."""

        bar = self.tabBar(row)
        index = self.stack.addWidget(page)

        tabIndex = bar.addTab(*args, **kwargs)
        bar.setTabData(tabIndex, page)

        if self.currentRow != row:
            bar.setCurrentIndex(0)
            #bar.setDrawBase(False)
        elif bar.count() == 2:
            bar.setCurrentIndex(1)

        return index

    def _relayout(self):
        """Re-order the widget layout, optionally creating it.
        This happens quite often as the current tab needs to be
        at the bottom of it."""

        if not self.layout():
            self.setLayout(QtWidgets.QVBoxLayout())
            self.layout().setContentsMargins(0, 0, 0, 0)
            self.layout().setSpacing(0)
        while self.layout().takeAt(0):
            pass

        for bar in reversed(self.tabBars[:self.currentRow]):
            self.layout().addWidget(bar)

        for bar in reversed(self.tabBars[self.currentRow:]):
            self.layout().addWidget(bar)

        self.layout().addWidget(self.stack, 1)


class FontInfoWidget(QtWidgets.QTabWidget):

    """Font information widget (dialog). Emits closed() signal
    so it can be used in TypeAtlas window list."""

    def __init__(self, item: 'typeatlas.fontlist.FontLike',
                       options: Options=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setTabPosition(self.West)

        if options is None:
            options = Options.getInstance()

        add = partial(_add_key_value_to_grid, options=options)
        add_separator = _add_separator_to_grid

        extended = item.extended()
        fontCmap = extended.cmap

        if fontCmap is not None:
            chars = sorted(fontCmap)
        elif hasattr(QtGui, 'QRawFont') and False:
            font = QtGui.QFontDatabase().font(item.family, item.style, 16)
            rawFont = QtGui.QRawFont.fromFont(font)
            allChars = all_character_codes()
            chars = list(filter(rawFont.supportsCharacter, allChars))
        else:
            chars = list(all_character_codes())

        charblocks = list(blockmath.toblocks_inorder(chars))

        fileinfos = list(item.files(details=True, stat=True))

        # ==== Tab 1 ====
        summary = QtWidgets.QWidget()
        self.addTab(summary, _('Summary'))
        summary.setLayout(QtWidgets.QVBoxLayout())

        title = QtWidgets.QHBoxLayout()
        summary.layout().addLayout(title)

        icon = QtWidgets.QLabel()
        label = QtWidgets.QLabel()

        title.addWidget(icon)
        title.addWidget(label, 1)
        icon.setPixmap(getIcon(item.icon).pixmap(options.tooltipIconSize))
        label.setTextFormat(Qt.PlainText)
        if item.is_family:
            titleText = item.translate('family', textlang())
        else:
            titleText = item.translate('fullname', textlang())
        self.setWindowTitle(titleText)
        self.setWindowIcon(getIcon('font-item-' +
                                   item.file_format_info.icon))
        label.setText(titleText)
        label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)

        grid = QtWidgets.QGridLayout()
        summary.layout().addLayout(grid)

        add(grid, _('Type'), _('Font style') if item.is_style else
                             _('Font family'),
            icon='font-item-style' if item.is_style else
                 'font-item-family')
        add(grid, _('Family'), item.translate('family', textlang()))
        add(grid, _('Style') if item.is_style else
                  _('Main style'), item.translate('style', textlang()))
        add(grid, _('Generic family'), _(item.genericfamily) or
                                       _('Unclassified'),
                                       icon=item.genericfamily)

        if item.panoseclass or item.ibmclass:
            add_separator(grid)

        if item.panoseclass:
            add(grid, _('PANOSE class'), _(item.panoseclass.class_name))
            add(grid, _('PANOSE subclass'), _(item.panoseclass.subclass_name))

        if item.ibmclass:
            add(grid, _('IBM class'), _(item.ibmclass.class_name))
            add(grid, _('IBM subclass'), _(item.ibmclass.subclass_name))

        add_separator(grid)


        fontFormat = item.font_format_info
        fileFormat = item.file_format_info

        more = None
        if fontFormat.annotation:
            moreWidget = AnnotationWidget(item, fontFormat, False,
                                          options=options)
            more = moreWidget.show
            self.formatMoreWidget = moreWidget
        add(grid, _('Font format'), fontFormat.name, fontFormat.icon,
            moreInfoSlot=more)

        more = None
        if fileFormat.annotation:
            moreWidget = AnnotationWidget(item, fileFormat, True,
                                          options=options)
            more = moreWidget.show
            self.fileMoreWidget = moreWidget
        add(grid, _('File format'), fileFormat.name, fileFormat.icon,
            moreInfoSlot=more)

        add_separator(grid)

        add(grid, _('Character count'), str(len(chars)))

        if item.is_family:
            add(grid, _('Style count'), str(len(item.styles)))

        if fileinfos:
            add(grid, _('Files'), str(len(fileinfos)))
            add(grid, _('Size'), format_size(sum(fi.stat.st_size
                                                 for fi in fileinfos
                                                 if fi.stat)))

        grid.addWidget(QtWidgets.QWidget(), grid.rowCount(), 0, 1, 3)
        grid.setColumnStretch(grid.columnCount() - 1, 1)
        grid.setRowStretch(grid.rowCount() - 1, 1)

        # ==== Tab 2 ====

        files = QtWidgets.QTableWidget()
        tableItem = QtWidgets.QTableWidgetItem
        self.addTab(files, _('Files'))

        if item.is_style:
            offset = 0
            files.setColumnCount(5)
        else:
            offset = 1
            files.setColumnCount(6)

        setResizeMode(files.horizontalHeader(),
                      QtWidgets.QHeaderView.ResizeToContents)

        #fileIcon = getIcon('font-x-generic')
        files.setHorizontalHeaderItem(0, tableItem(_('File')))
        files.setHorizontalHeaderItem(1, tableItem(_('No.')))
        files.setHorizontalHeaderItem(2, tableItem(_('Size')))
        if not item.is_style:
            files.setHorizontalHeaderItem(3, tableItem(_('Style')))
        files.setHorizontalHeaderItem(3 + offset, tableItem(_('Purpose')))
        files.setHorizontalHeaderItem(4 + offset, tableItem(_('Directory')))

        for i, fi in enumerate(fileinfos):

            if fi.metrics:
                fileIcon = getIcon('font-file-metrics')
            else:
                fileIcon = getIcon('font-file-' +
                                   fi.style.file_format_info.icon)

            files.insertRow(i)
            files.setItem(i, 0, tableItem(fileIcon, os.path.basename(fi.file)))
            files.setItem(i, 1, tableItem(str(fi.index)))
            if fi.stat:
                files.setItem(i, 2, tableItem(format_size(fi.stat.st_size)))
            if not item.is_style:
                styleName = fi.style.translate('style', textlang())
                files.setItem(i, 3, tableItem(styleName))
            files.setItem(i, 3 + offset, tableItem(_(fi.purpose)))
            files.setItem(i, 4 + offset, tableItem(os.path.dirname(fi.file)))

        # ==== Tab 3 ====
        if item.panoseclass:
            panose = QtWidgets.QWidget()
            self.addTab(panose, _('PANOSE information'))
            grid = QtWidgets.QGridLayout()
            panose.setLayout(grid)

            for panprop in item.panoseclass:
                add(grid, _(panprop.fieldname), panprop.text(translate=_))

            grid.addWidget(QtWidgets.QWidget(), grid.rowCount(), 0, 1, 3)
            grid.setColumnStretch(grid.columnCount() - 1, 1)
            grid.setRowStretch(grid.rowCount() - 1, 1)

    closed = Signal()

    def closeEvent(self, event):
        r = super().closeEvent(event )
        if event.isAccepted():
            self.closed.emit()
        return r


class CharInfoWidget(QtWidgets.QTabWidget):

    """Character information widget. Emits closed() signal
    so it can be used in TypeAtlas window list."""

    def __init__(self, fontItem: 'typeatlas.fontlist.FontLike',
                       char: int, options: Options=None,
                       charDb: CharacterDatabase=None,
                       fontDb: QtGui.QFontDatabase=None,
                       langDb: LanguageDatabase=None,
                       *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setTabPosition(self.West)

        if options is None:
            options = Options.getInstance()
        if charDb is None:
            charDb = CharacterDatabase.getInstance()
        if langDb is None:
            langDb = LanguageDatabase.getInstance()
        if fontDb is None:
            fontDb = QtGui.QFontDatabase()

        add = partial(_add_key_value_to_grid, options=options)
        add_separator = _add_separator_to_grid

        font = fontDb.font(fontItem.family, fontItem.style,
                           options.gridFontSize)

        info = charDb.get(char, autofill=True)

        plane = charDb.get_plane(char)
        script = charDb.find_script_name(langDb, char)
        symbol = chr(char)
        display = symbol
        if getattr(info, 'display', None):
            display = chr(info.display)

        extended = fontItem.extended()
        if extended and extended.cmap:
            fontCmap = extended.cmap
            fontCharName = fontCmap.get(char) or ''
        else:
            fontCmap = None
            fontCharName = ''

        titleText = (getattr(info, 'name', '') or
                     getattr(info, 'old_name', '') or
                     fontCharName or '0x%x' % (char, ))

        self.setWindowTitle(titleText.title())
        self.setWindowIcon(getIcon(info.category_icon))

        # ==== Tab 1 ====
        summary = QtWidgets.QWidget()
        self.addTab(summary, _('Summary'))
        summary.setLayout(QtWidgets.QVBoxLayout())

        title = QtWidgets.QHBoxLayout()
        summary.layout().addLayout(title)

        icon = QtWidgets.QLabel()
        label = QtWidgets.QLabel()

        title.addWidget(icon)
        title.addWidget(label, 1)
        icon.setFont(font)
        icon.setTextFormat(Qt.PlainText)
        icon.setText(display)

        label.setTextFormat(Qt.PlainText)
        label.setText(titleText)
        label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)

        grid = QtWidgets.QGridLayout()
        summary.layout().addLayout(grid)

        add_separator(grid)

        if info:
            if info.name:
                add(grid, _('Unicode name'), info.name)
            if info.old_name:
                add(grid, _('Old unicode name'), info.old_name)

        if fontCharName:
            add(grid, _('Character name in font'), fontCharName)
        add(grid, _('In font'), fontItem.translate('fullname', textlang()))

        add_separator(grid)


        if info and info.block:
            add(grid, _('Block'), info.block.name)
        if script:
            add(grid, _('Script'), script)
        if plane:
            add(grid, _('Plane'), _('Plane %d') % (plane.number, ) + ' ' +
                                   _(plane.description))

        add_separator(grid)

        add(grid, _('Hex code'), '0x%x' % (char, ))
        add(grid, _('Character'), display)
        if info:
            add(grid, _('Category'), info.category_name(translate=_),
                                     icon=info.category_icon)

        text = []
        if info:
            for alias in info.aliases:
                text.append('= ' + alias)
            for alias in info.formalaliases:
                text.append(charinfo.formal_alias + ' ' + alias)

        if text:
            aliases = QtWidgets.QGroupBox(_('Aliases'))
            summary.layout().addWidget(aliases)
            aliases.setLayout(QtWidgets.QVBoxLayout())
            aliasText = QtWidgets.QLabel()
            aliases.layout().addWidget(aliasText)
            aliasText.setTextFormat(Qt.PlainText)
            aliasText.setText('\n'.join(text))

        grid.addWidget(QtWidgets.QWidget(), grid.rowCount(), 0, 1, 3)
        grid.setColumnStretch(grid.columnCount() - 1, 1)
        grid.setRowStretch(grid.rowCount() - 1, 1)

        # ==== Tab 2 ====
        if info:
            for variation in info.variations:
                text.append(_('Variation', ': ') + variation)
            for crossref in info.crossrefs:
                text.append(charinfo.cross_reference + ' ' + crossref)

        if text:
            crossrefs = QtWidgets.QTextEdit()
            self.addTab(crossrefs, _('Names and references'))
            crossrefs.setPlainText('\n'.join(text))

    closed = Signal()

    def closeEvent(self, event):
        r = super().closeEvent(event)
        if event.isAccepted():
            self.closed.emit()
        return r



GROUP = N_('group')

EDIT = N_('edit')
CREATE = N_('create')


class GroupNameDialog(QtWidgets.QDialog):

    """A dialog for creating and editing groups (categories, tags,
    searches).

    Providing custom arguments, it can be used with any items having
    a name, icon and color, such as interface configurations. Then
    you need to provide groupTypeName, and the optional list of
    items as availableGroups.

    If nameEditable=False is passed, edint the name is disabled.
    """

    def __init__(self, *args,
                 groupContainer: 'typeatlas.datastore.FontGroupsContainer'=None,
                 action: str=CREATE, groupTypeName: str=None,
                 availableGroups: Sequence=None,
                 replacesOldGroup: bool=False,
                 nameEditable: bool=True,
                 groupName: str=None,
                 **kwargs):

        super().__init__(*args, **kwargs)

        if groupTypeName is None:
            if groupContainer is not None:
                groupTypeName = groupContainer.factory.type_label
            else:
                groupTypeName = GROUP

        if availableGroups is None:
            if groupContainer is not None:
                availableGroups = list(groupContainer.info.values())
            else:
                availableGroups = ()

        # TODO: For searches, check that name already exists
        # TODO: Signals here, and in the filter widget for search saving/editing
        # TODO: Edit is the same as replace, and it should be different
        self.buttons = buttons = QtWidgets.QDialogButtonBox()

        self.color = None
        self.icon = None
        self.replacesOldGroup = replacesOldGroup

        self.groupTypeName = groupTypeName

        self.setWindowTitle(_('Save {}').format(_(groupTypeName)))

        buttons.addButton(_('&Save'), buttons.AcceptRole)
        buttons.addButton(_('&Cancel'), buttons.RejectRole)

        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        self.nameCombo = combo = QtWidgets.QComboBox()
        self.nameCombo.setEditable(True)

        self.availableGroups = {}
        for i, grouping in enumerate(availableGroups):
            if grouping.icon:
                combo.addItem(getIcon(grouping.icon), grouping.name, grouping)
            else:
                combo.addItem(grouping.name, grouping)

            self.availableGroups[grouping.name] = i, grouping

        self.nameCombo.setCurrentIndex(-1)
        self.nameCombo.clearEditText()

        combo.currentIndexChanged.connect(self.nameIndexChanged)
        combo.editTextChanged.connect(self.nameTextChanged)

        self.grouping = None

        self.iconPreview = QtWidgets.QLabel()
        self.colorPreview = QtWidgets.QLabel()

        self.iconButton = QtWidgets.QPushButton(_("Icon..."))
        self.iconButton.clicked.connect(self.selectIcon)
        self.fontIconButton = QtWidgets.QPushButton("Symbol...")
        self.fontIconButton.clicked.connect(self.selectFontIcon)
        self.iconClear = QtWidgets.QPushButton(_("Clear"))
        self.iconClear.clicked.connect(self.setIcon)

        self.colorButton = QtWidgets.QPushButton(_("Color..."))
        self.colorButton.clicked.connect(self.selectColor)
        self.colorClear = QtWidgets.QPushButton(_("Clear"))
        self.colorClear.clicked.connect(self.setColor)

        self.nameCombo.lineEdit().setPlaceholderText(_('Name...'))

        self.setLayout(QtWidgets.QGridLayout())
        self.layout().addWidget(self.nameCombo, 0, 0, 1, 7)

        self.layout().addWidget(self.iconPreview, 1, 0, 1, 1)
        self.layout().addWidget(self.iconButton, 1, 1, 1, 1)
        self.layout().addWidget(self.fontIconButton, 1, 2, 1, 1)
        self.layout().addWidget(self.iconClear, 1, 3, 1, 1)

        self.layout().addWidget(self.colorPreview, 1, 4, 1, 1)
        self.layout().addWidget(self.colorButton, 1, 5, 1, 1)
        self.layout().addWidget(self.colorClear, 1, 6, 1, 1)

        self.layout().addWidget(self.buttons, 2, 0, 1, 7)

        if groupName is not None:
            self.setName(groupName)

        if not nameEditable:
            self.nameCombo.setDisabled(True)

    nameChanged = Signal(str)
    iconChanged = Signal(str)
    colorChanged = Signal(QtGui.QColor)

    @property
    def groupName(self) -> str:
        """The name of the group (category, tag, search) or object
        being edited."""
        return self.nameCombo.currentText()

    @Slot()
    @Slot(str)
    def setName(self, name: str=''):
        """Set the name."""
        if name == self.nameCombo.currentText():
            return
        if not name:
            self.nameCombo.clearEditText()
        else:
            self.nameCombo.setEditText(name)
        self.nameChanged.emit(name)

    @Slot()
    @Slot(str)
    def setIcon(self, path: str=None):
        """Set the path to the icon. It can be taken from self.icon"""
        if path == self.icon:
            return
        if path is None:
            self.iconPreview.clear()
        else:
            icon = getIcon(path)
            self.iconPreview.setPixmap(icon.pixmap(iconSize(refSize=32)))

        self.icon = path
        self.iconChanged.emit(path)

    @Slot()
    @Slot(QtGui.QColor)
    def setColor(self, color: QtGui.QColor=None):
        """Set the color. It can be taken from self.color, or as string
        with self.colorName()."""
        if color == self.color:
            return
        if color is None:
            self.colorPreview.clear()
        else:
            pixmap = QtGui.QPixmap(iconSize(refSize=32))
            pixmap.fill(color)
            self.colorPreview.setPixmap(pixmap)
        self.color = color
        self.colorChanged.emit(color)

    def colorName(self) -> Optional[str]:
        """Get the color as a string. None if unset."""
        if self.color is None:
            return None
        return self.color.name()

    @Slot()
    def selectColor(self):
        """Open the color selection dialog"""
        color = QtWidgets.QColorDialog.getColor(
                        QtGui.QColor(), self,
                        _('Select %s color') % (_(self.groupTypeName), ))
        if color:
            self.setColor(color)

    @Slot()
    def selectIcon(self):
        """Open the icon selection dialog"""
        icon = IconSelectDialog.selectIcon(self, _('Select %s icon') % (
                                                        _(self.groupTypeName), ))
        if icon:
            self.setIcon(icon)

    @Slot()
    def selectFontIcon(self):
        """Open the emoji icon selection dialog"""
        from typeatlas.guiglyphs import CharSelectDialog
        charRes = CharSelectDialog.selectChar()
        if charRes is None:
            return

        if charRes.generic:
            self.setIcon('char:U+%X' % (charRes.codepoint, ))
        else:
            self.setIcon('char:U+%X:%s' % (charRes.codepoint,
                                           charRes.fontitem.family))

    @Slot(int)
    def nameIndexChanged(self, idx: int):
        """A combo item was selected. We take its name, color and
        icon."""
        grouping = self.nameCombo.itemData(idx)
        if grouping is None:
            return

        self.grouping = grouping
        self.setName(grouping.name)
        self.setIcon(grouping.icon)
        if grouping.color:
            self.setColor(QtGui.QColor(grouping.color))
        else:
            self.setColor()

    @Slot(str)
    def nameTextChanged(self, name: str):
        """The name was selected. If the item exists, take its icon and
        color."""
        groupingTuple = self.availableGroups.get(name)
        if groupingTuple is None:
            return

        i, grouping = groupingTuple
        self.setIcon(grouping.icon)
        if grouping.color:
            self.setColor(QtGui.QColor(grouping.color))
        else:
            self.setColor()
        self.nameCombo.setCurrentIndex(i)


class RemoveFromGroupDialog(QtWidgets.QDialog):

    """A dialog for font items from groups."""

    def __init__(self,
                 families: 'IterableOf[str]',
                 groupContainer: 'typeatlas.datastore.FontGroupsContainer',
                 *args, **kwargs):

        super().__init__(*args, **kwargs)

        groupTypeName = groupContainer.factory.type_label
        groupTypePlural = groupContainer.factory.type_label_plural
        #availableGroups = list(groupContainer.info.values())

        self.setWindowTitle(_('Remove {} for selected the fonts')
                                .format(_(groupTypePlural)))

        self.buttons = buttons = QtWidgets.QDialogButtonBox()

        buttons.addButton(_('&Remove'), buttons.AcceptRole)
        buttons.addButton(_('&Cancel'), buttons.RejectRole)

        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        self.groups = groups = set()
        self.sharedGroups = groupCounts = Counter()
        self.groupCounts = sharedGroups = None

        for family in families:
            familyGroups = groupContainer.by_family.get(family, ())
            if sharedGroups is None:
                sharedGroups = set(familyGroups)
            else:
                sharedGroups &= familyGroups

            groups |= familyGroups

            groupCounts.update(familyGroups)

        if sharedGroups is None:
            sharedGroups = set()

        self.tree = tree = QtWidgets.QTreeWidget()
        tree.setColumnCount(3)
        tree.setHeaderLabels([_(groupTypeName).capitalize(),
                              _('Fonts'), _('Count')])
        tree.setSelectionMode(tree.ExtendedSelection)

        for group in sorted(groups, key=groupCounts.get, reverse=True):
            item = QtWidgets.QTreeWidgetItem(
                [group,
                 _('All') if group in sharedGroups else _('Some'),
                 str(groupCounts[group])])

            icon = groupContainer.info[group].icon
            if icon:
                icon = getIcon(icon)
                if icon:
                    item.setIcon(0, icon)

            tree.addTopLevelItem(item)

        ## Can't search in QTreeWidget, unfortunately.
        #self.textWidget = search = QuickSearch(
        #    placeholderText=_('Search {}...').format(_(groupTypePlural)))
        #search.searchTriggered.connect(self._filterTextChanged)

        self.setLayout(QtWidgets.QVBoxLayout())

        self.layout().addWidget(tree)
        self.layout().addWidget(buttons)

    def selectedGroups(self) -> SequenceOf[str]:
        """Return a list of the selected groups."""
        selected = self.tree.selectionModel().selectedIndexes()
        if not selected:
            current = self.tree.selectionModel().currentIndex()
            if current.isValid():
                selected = [current]

        # We have multiple columns selected, but only need the first one,
        # and need it returned only once.
        return list(OrderedSet(index.siblingAtColumn(0).data(Qt.DisplayRole)
                               for index in selected))

    #@Slot(str)
    #def _filterTextChanged(self, text):
    #    self.proxyModel.setFilterFixedString(text.strip())


class IconSelectDialog(QtWidgets.QDialog):

    """A widget for selecting icons."""

    def __init__(self, *args,  **kwargs):
        super().__init__(*args, **kwargs)
        
        self.iconsByCategory = OrderedDict()

        groupAttr = attrgetter('context')

        for context, iconInfos in groupby(sorted(osinfo.find_icons(),
                                                 key=groupAttr), groupAttr):
            self.iconsByCategory[context] = list(iconInfos)

        self.buttons = buttons = QtWidgets.QDialogButtonBox()

        buttons.addButton(_('&OK'), buttons.AcceptRole)
        buttons.addButton(_('&Cancel'), buttons.RejectRole)
        self.browseButton = buttons.addButton(_('&Browse...'), buttons.ActionRole)

        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        buttons.clicked.connect(self._buttonClicked)

        self.listWidget = QtWidgets.QListWidget()
        self.comboBox = QtWidgets.QComboBox()

        self.setLayout(QtWidgets.QVBoxLayout())

        self.layout().addWidget(self.comboBox)
        self.layout().addWidget(self.listWidget)
        self.layout().addWidget(self.buttons)

        self.comboBox.addItems(list(self.iconsByCategory))
        self.comboBox.activated.connect(self._comboActivated)

        self.listWidget.setViewMode(self.listWidget.IconMode)

        size = iconSize(refSize=32)
        self.listWidget.setIconSize(size)
        self.listWidget.setGridSize(size * 2)

        self.listWidget.setMovement(self.listWidget.Static)
        self.listWidget.setLayoutMode(self.listWidget.Batched)
        self.listWidget.currentItemChanged.connect(self._iconChanged)
        self.listWidget.setResizeMode(self.listWidget.Adjust)
        self.listWidget.setWordWrap(True)
        self.listWidget.setSortingEnabled(True)

        if self.iconsByCategory:
            self.comboBox.setCurrentIndex(0)
            self._comboActivated(0)

        self.resize(generalWidth(refSize=500),
                    generalHeight(refSize=600))

        self.result = None

    @classmethod
    def selectIcon(cls, parent: QtWidgets.QWidget=None,
                        title: str=None) -> Optional[str]:
        """Execute the dialog in modal mode and return icon name."""
        dialog = cls(parent)
        dialog.setWindowTitle(title)
        if dialog.exec_():
            return dialog.result
        return None

    @Slot(QtWidgets.QAbstractButton)
    def _buttonClicked(self, button: QtWidgets.QAbstractButton):
        """A button in the button box was clicked. If it is the browse
        button, open the a file dialog to select an image."""
        if button is self.browseButton:
            (path, nameFilter) = QtWidgets.QFileDialog.getOpenFileName(
                                    self, self.windowTitle())
            if path:
                self.result = path
                self.accept()

    @Slot(int)
    def _comboActivated(self, index: int):
        """Combo was selected, move to a different font category."""
        category = self.comboBox.itemText(index)
        self.listWidget.clear()

        for iconItem in self.iconsByCategory.get(category, ()):
            name = iconItem.name
            icon = getIcon(name)
            self.listWidget.addItem(QtWidgets.QListWidgetItem(icon, name))

    @Slot(QtWidgets.QListWidgetItem, QtWidgets.QListWidgetItem)
    def _iconChanged(self, current: QtWidgets.QListWidgetItem,
                           previous: QtWidgets.QListWidgetItem):
        """The icon was changed, update the result."""
        self.result = current.text()




# -*- coding: utf-8 -*-
#
#    TypeAtlas Font List implemented using the Qt Library
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import absolute_import, division

import os
import re
import struct
import os.path
import typeatlas.fontlist as fontlist
from collections import defaultdict
from collections.abc import Mapping, Set
from typeatlas import opentype, blockmath, rangemath, charinfo, event
from typeatlas.util import OrderedSet, generic_type
from typeatlas.langutil import _
from typeatlas.compat import QtCore, QtGui, QtWidgets, Qt, Slot, Signal
from typeatlas.compat import qtGetBytes
from itertools import chain, count


SetOf = generic_type('Set')
IterableOf = generic_type('Iterable')
IteratorOf = generic_type('Iterator')
Optional = generic_type('Optional')


# TODO: Can't translate the styles here?

# TODO: Loading info is quite slow due to opening all SFNT tables,
#       and all fonts by file without cache. Maybe allow skipping this?


def getWeight(font: QtGui.QFont) -> float:
    """Get the weight of the QFont in the same scale fontconfig would
    use."""

    # FIXME: PySide constants suck, only work on the right-hand side
    # of an equation (broken __rsub__), so QtCore.Qt.UserRole + 3 is OK,
    # 3 + QtCore.Qt.UserRole is not (same for QtGui.QFont.Normal)
    # tag: zjsZ0zzunjo

    factor = ((fontlist.WEIGHT_BOLD - fontlist.WEIGHT_NORMAL) /
                (QtGui.QFont.Bold - QtGui.QFont.Normal))
    return ((font.weight() - int(QtGui.QFont.Normal)) * factor
                    + fontlist.WEIGHT_NORMAL)


def getWidth(font: QtGui.QFont) -> float:
    """Get the width of the QFont in the same scale fontconfig would
    use."""

    # FIXME: see tag: zjsZ0zzunjo
    factor = ((fontlist.WIDTH_NORMAL - fontlist.WIDTH_CONDENSED) /
                (QtGui.QFont.Unstretched - QtGui.QFont.Condensed))
    return ((font.stretch() - int(QtGui.QFont.Unstretched)) * factor
                    + fontlist.WIDTH_NORMAL)


def getSlant(font: QtGui.QFont) -> int:
    """Get the width of the QFont in the same scale fontconfig would
    use."""
    style = font.style()
    if style == QtGui.QFont.StyleNormal:
        return fontlist.SLANT_NORMAL
    elif style == QtGui.QFont.StyleItalic:
        return fontlist.SLANT_ITALIC
    elif style == QtGui.QFont.StyleOblique:
        return fontlist.SLANT_OBLIQUE


def getFont(fontDb: QtGui.QFontDatabase,
            familyName: str, styleName: str,
            fileinfos: IterableOf[fontlist.FontFileDetectInfo]=()
            ) -> fontlist.Font:
    """Get the font information about the given family and style from the
    Qt font database, the provided file infos."""

    font = fontlist.Font()

    font.family = familyName
    font.style = styleName

    qfont = fontDb.font(familyName, styleName, 16)

    font.weight = getWeight(qfont)
    font.slant = getSlant(qfont)
    font.width = getWidth(qfont)

    font.outline = fontDb.isSmoothlyScalable(familyName, styleName)
    font.scalable = fontDb.isScalable(familyName, styleName)

    font.file = getFileName(qfont)

    font.set_cachekey(familyName + ' ' + styleName)

    rawfont = None

    fileinfos = list(fileinfos)

    if fileinfos:
        font.file_alts = [fontlist.FontFile(fi.file, fi.index)
                          for fi in fileinfos]
        font.file, font.index, formathint = fileinfos[0]

        if formathint == 'SFNT':
            font.sfnt = True
            pass

        elif formathint:
            font.fontformat = formathint

        elif font.outline or font.scalable:
            font.fontformat = 'Type 1'

    elif hasattr(QtGui, 'QRawFont'):
        rawfont = QtGui.QRawFont.fromFont(qfont)
        if rawfont.fontTable('loca'):
            font.fontformat = 'TrueType'
            font.sfnt = True
            if not font.file:
                font.ext = 'ttf'
        elif rawfont.fontTable('CFF ') or rawfont.fontTable('CFF2'):
            if not font.file:
                font.ext = 'otf'
            font.fontformat = 'CFF'
            font.sfnt = True
        elif (not rawfont.fontTable('head') and
                (font.outline or font.scalable)):
            if not font.file:
                # That's bad guesswork. It may not be true.
                # If this does not get hit, fontformat will be filled from
                # extension. Type 1 fonts are not multi-format, so extension
                # is good enough (probably better than this over-optimistic
                # guess).
                font.fontformat = 'Type 1'
                font.ext = 't1'

    #fillDetailedInfo(font, qfont, rawfont)
    fontlist.autofill_font_info(font)

    return font


def updateFamilyInfo(fontDb: QtGui.QFontDatabase, family: fontlist.FontFamily):
    """Fill any family information, particularly writing system
    and generic family. This is to be called manually at the moment."""

    family.writingSystems = fontDb.writingSystems(family.family)
    for style in family.styles:
        style.writingSystems = family.writingSystems
    if list(family.writingSystems) == [QtGui.QFontDatabase.Symbol]:
        for style in family.styles:
            style.genericfamily = 'symbol'
            style.genericfamily_source = fontlist.SOURCE_QT


def _fillDetailedInfo(font: fontlist.Font, qfont: QtGui.QFont,
                      rawfont: 'QtGui.QRawFont'=None):
    """Fill any detailed information about the font, such as PANOSE
    class. This is the actual implementation called
    by fillDetailedInfo.
    """
    if hasattr(font, 'panoseclass'):
        return

    # FIXME: This fills up memory and never frees it.
    if rawfont is None:
        if not hasattr(QtGui, 'QRawFont'):
            font.ibmclass = opentype.NO_IBM_CLASS_DATA
            font.panoseclass = opentype.NO_PANOSE_DATA
            font.embedding = opentype.NO_EMBEDDING_INFO
            return
        rawfont = QtGui.QRawFont.fromFont(qfont)

    ## NOTE This is extremely slow, don't do it.
    #allChars = charinfo.all_character_codes()
    #chars = filter(rawfont.supportsCharacter, allChars)
    #font.charblocks = list(blockmath.toblocks_inorder(chars))

    os2 = rawfont.fontTable('OS/2')
    if not os2:
        font.ibmclass = opentype.NO_IBM_CLASS_DATA
        font.panoseclass = opentype.NO_PANOSE_DATA
        font.embedding = opentype.NO_EMBEDDING_INFO
        return

    os2 = qtGetBytes(os2)

    font.ibmclass = opentype.extract_ibm_class(os2)
    font.panoseclass = opentype.extract_panose_class(os2)
    font.embedding = opentype.extract_embedding_info(os2)
    if not font.genericfamily:
        font.genericfamily = font.panoseclass.genericfamily
        font.genericfamily_source = fontlist.SOURCE_PANOSE
    if not font.genericfamily:
        font.genericfamily = font.ibmclass.genericfamily
        font.genericfamily_source = fontlist.SOURCE_IBM


class _Filler(QtCore.QThread):
    """A thread that fills the detailed font info with
    _fillDetailedInfo. This is not presently being used."""

    def __init__(self, font: fontlist.Font, qfont: QtGui.QFont):
        super(_Filler, self).__init__()
        self.font = font
        self.qfont = qfont

    def run(self):
        _fillDetailedInfo(self.font, self.qfont)


def fillDetailedInfo(font: fontlist.Font, qfont: QtGui.QFont,
                     rawfont: 'QtGui.QRawFont'=None):
    """Fill any detailed information about the font, such as PANOSE
    class. This is to be called when fontlist.fill_detailed_info
    raises a fontlist.NotSupportedError only, which means fontTools
    was not available to fill it.
    """
    _fillDetailedInfo(font, qfont, rawfont)
    return

    if hasattr(font, 'panoseclass'):
        return
    if rawfont is not None:
        return _fillDetailedInfo(font, qfont, rawfont)

    thread = _Filler(font, qfont)
    thread.start()
    thread.wait()
    thread.deleteLater()


def getFileName(font: QtGui.QFont) -> str:
    """Get the file name of the font? No, can't do. A place holder for
    when we eventually figure out a way."""
    try:
        # This doesn't really work, resort to other forms of discovery.
        #return font.rawName()
        return ''
    except AttributeError:
        return ''


class _StopCrawling(Exception):
    """Stop crawling font files - internal exception."""


class QtFontFinder(fontlist.FontFinder):

    def __init__(self, fontDb: QtGui.QFontDatabase=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if fontDb is None:
            fontDb = QtGui.QFontDatabase()
        self.fontDb = fontDb

    def _findFontFiles(self, knownFamilies: Set=()
                       ) -> fontlist.FamilyFilePathType:
        """Find the font files associated with family and style names.
        This tries to use fontlist.crawl_font_directories() if fontTools
        is available, if that's not supported, this uses the slow code path
        which uses QRawFont to fill the information.
        """
        locations = self.fontdirs()

        i = 0
        abort = False
        autoabort = True

        fontnum = count(0)

        def progress(num: int=0):
            """Tick the progress."""
            self.fontprogress(
                    num, len(knownFamilies),
                    message=_('First run: Discovering font files: '
                              '{} out of {}').format(num, len(knownFamilies)))

        def familyAdded(name: str, ignored: Mapping):
            """A family was added to the font files dictionary. Tick the
            progress bar."""
            nonlocal i, abort

            if name not in knownFamilies:
                return

            i += 1
            progress(i)

            if (fontlist.FONT_LIMIT is not None and
                next(fontnum) >= fontlist.FONT_LIMIT):

                abort = True
                if autoabort:
                    # FIXME: Our API allows this, but it's not documented.
                    #        Probably because it's ugly.
                    #        We only get here in debug mode, so who cares?
                    def circuitbreaker(*args, **kwargs):
                        raise _StopCrawling()
                    fontFiles.added = circuitbreaker

        fontFiles = event.NoisyMapping()
        fontCounts = {}

        fontFiles.added.connect(familyAdded)
        progress()

        try:
            fontlist.crawl_font_directories(locations, fontFiles, fontCounts)
        except fontlist.NotSupportedError:
            pass
        except _StopCrawling:
            pass

        autoabort = False

        if not hasattr(QtGui, 'QRawFont') or abort:
            self.fontprogressdone()
            return fontFiles

        for location in locations:
            if abort:
                break

            for dirpath, dirnames, filenames in os.walk(location):
                if abort:
                    break

                for filename in filenames:
                    filepath = os.path.join(dirpath, filename)
                    if fontCounts.get(filepath):
                        continue

                    rawFont = QtGui.QRawFont(filepath, 16)

                    family = rawFont.familyName()
                    style = rawFont.styleName()

                    if family in fontFiles:
                        styleFiles = fontFiles[family]
                    else:
                        styleFiles = fontFiles[family] = defaultdict(OrderedSet)

                    if rawFont.fontTable('loca'):
                        formathint = 'TrueType'

                    elif rawFont.fontTable('CFF ') or rawFont.fontTable('CFF2'):
                        formathint = 'CFF'

                    elif rawFont.fontTable('head'):
                        formathint = 'SFNT'

                    else:
                        formathint = None

                    fontCounts[filepath] = 1
                    styleFiles[style].add(fontlist.FontFileDetectInfo(
                                                filepath, 0, formathint))


        self.fontprogressdone()
        return fontFiles

    def fontdirs(self) -> SetOf[str]:
        fontdirs = OrderedSet()
        if hasattr(QtCore, 'QStandardPaths'):
            fontdirs.update(QtCore.QStandardPaths.standardLocations(
                                QtCore.QStandardPaths.FontsLocation))
        else:
            fontdirs.add(QtGui.QDesktopServices.storageLocation(
                                QtGui.QDesktopServices.FontsLocation))
        try:
            fontdirs.update(super().fontdirs())
        except fontlist.NotSupportedError:
            pass
        return fontdirs

    @classmethod
    def supported(self) -> bool:
        return True

    @classmethod
    def remote_supported(self) -> bool:
        return False

    def fonts(self) -> IteratorOf[fontlist.Font]:

        try:
            yield from super().fonts()
        except fontlist.NotSupportedError:
            pass
        else:
            return

        cache = self.metadata_cache

        fontDb = self.fontDb
        familyNames = OrderedSet(fontDb.families())
        familyStyles = {}

        crawlAllFiles = False

        for familyName in familyNames:
            familyStyles[familyName] = fontDb.styles(familyName) or ['Regular']

        # Attempting to fill info with cached font files.
        cachedFiles = {}

        if cache is None:
            crawlAllFiles = True

        else:
            # Count fonts for FONT_LIMIT
            fontnum = count()

            for familyName in familyNames:

                if (fontlist.FONT_LIMIT is not None and
                    next(fontnum) >= fontlist.FONT_LIMIT):

                    crawlAllFiles = False
                    break

                for styleName in familyStyles[familyName]:
                    cachekey = familyName + ' ' + styleName
                    fileinfos = cache.get_field(cachekey, 'fileinfos')
                    if fileinfos is not None:
                        try:
                            cachedFiles[familyName, styleName] = [
                                fontlist.FontFileDetectInfo(*info)
                                for info in fileinfos
                            ]
                        except (TypeError, ValueError) as exc:
                            cache.field_corrupt(cachekey, 'fileinfos', exc)

                    else:
                        crawlAllFiles = True
                        break

                if crawlAllFiles:
                    break

        if not crawlAllFiles:
            for familyName in familyNames:
                for styleName in familyStyles[familyName]:
                    fileinfos = cachedFiles[familyName, styleName]
                    font = getFont(fontDb, familyName, styleName, fileinfos)
                    font.source = 'qt'
                    font.finder = self
                    yield font
            return


        # If cached files fail us, attempt to find all font files,
        # create fonts that way, and cache those.
        fontFiles = self._findFontFiles(familyNames)

        for familyName in familyNames:
            styleFiles = fontFiles.get(familyName)
            if not styleFiles:
                # Qt tends to add foundry, like Zilla Slab [NONE],
                # we could alternatively fill in the foundry from the
                # first semi-colon column from opentype.UNIQUE_NAME_ID,
                # but why make it so complicated anyhoo.
                familyNameSimple = re.sub(r'\s+\[[^\]]+\]$', '', familyName)
                if familyNameSimple not in familyNames:
                    styleFiles = fontFiles.get(familyNameSimple)
                    if not styleFiles:
                        if fontlist.FONT_LIMIT is not None:
                            continue

            for styleName in familyStyles[familyName]:

                filenames = []

                if styleFiles:
                    fileinfos = styleFiles.get(styleName, ())
                else:
                    fileinfos = ()

                if not fileinfos:
                    if fontlist.FONT_LIMIT is not None:
                        continue

                font = getFont(fontDb, familyName, styleName, fileinfos)
                font.source = 'qt'
                font.finder = self

                if cache is not None:
                    cache.set_field(font, 'fileinfos', list(fileinfos))

                yield font

    def substitute(self, family: str, style: str) -> fontlist.Font:
        try:
            return super().substitute(family, style)
        except fontlist.NotSupportedError:
            pass

        return getFont(self.fontDb, family, style, ())

    def fill_detailed_info(self, font: fontlist.Font,
                                 qfont: QtGui.QFont=None,
                                 rawfont: 'QtGui.QRawFont'=None):
        if font.monospace is None:
            font.monospace = self.fontDb.isFixedPitch(font.family, font.style)

        try:
            return super().fill_detailed_info(font)
        except fontlist.NotSupportedError:
            pass

        if hasattr(font, 'panoseclass'):
            return

        if self.metadata_cache is not None:
            if self.metadata_cache.fill_classes(font):
                return

        if qfont is None:
            qfont = self.fontDb.font(font.family, font.style, 16)

        fillDetailedInfo(font, qfont, rawfont)

        if (self.metadata_cache is not None and
            font.panoseclass is not opentype.NO_PANOSE_DATA and
            font.ibmclass is not opentype.NO_IBM_CLASS_DATA):
                self.metadata_cache.cache_classes(font)

    def fill_charset(self, font: fontlist.Font):
        super().fill_charset(font)
        if font.charset is not None:
            return

        if not hasattr(QtGui, 'QRawFont'):
            return False

        # WARNING This is extremely extremely slow
        qfont = self.fontDb.font(font.family, font.style, 16)
        rawfont = QtGui.QRawFont.fromFont(qfont)
        allChars = charinfo.all_character_codes()
        chars = filter(rawfont.supportsCharacter, allChars)
        #font.charblocks = list(blockmath.toblocks_inorder(chars))
        font.charset = rangemath.OrdinalRange.from_iterable(chars)
        if self.metadata_cache is not None:
            self.metadata_cache.cache_charset(font)

        return True


_wsToKey = None
_keyToWs = None

def _populateWritingSystemMaps():
    """Populate writing system to name and name to writing
    system dictionaries."""

    global _wsToKey, _keyToWs

    _wsToKey = {}
    _keyToWs = {}

    for key, value in vars(QtGui.QFontDatabase).items():
        if not isinstance(value, QtGui.QFontDatabase.WritingSystem):
            continue
        key = str(key)
        _wsToKey[value] = key
        _keyToWs[key] = value


def writingSystemToString(ws: int, fallback: bool=True) -> str:
    """Return a string to serialize a QtGui.QFontDatabase.WritingSystem."""
    if _wsToKey is None:
        _populateWritingSystemMaps()
    try:
        return _wsToKey[ws]
    except KeyError:
        if not fallback:
            raise
        return str(ws)


def stringToWritingSystem(s: str, fallback: bool=True) -> int:
    """Return QtGui.QFontDatabase.WritingSystem from a string for
    serialization."""
    if _keyToWs is None:
        _populateWritingSystemMaps()

    try:
        return _keyToWs[s]
    except KeyError:
        if not fallback:
            raise
        s = str(s)
        if s.isdigit():
            return int(s)
        return s


_writingSystemAliases = None

def writingSystemAliases() -> Mapping:

    """Return writing system aliases of various kinds.
    Read the source. This is work in progress and not yet used."""

    global _writingSystemAliases

    if _writingSystemAliases is not None:
        return _writingSystemAliases

    scripts = {
        #'Any': '',
        'Latin': 'Latn',
        'Greek': 'Grek',
        'Cyrillic': 'Cyrl',
        'Armenian': 'Armn',
        'Hebrew': 'Hebr',
        'Arabic': 'Arab',
        'Syriac': 'Syrc',
        'Thaana': 'Thaa',
        'Devanagari': 'Deva',
        'Bengali': 'Beng',
        'Gurmukhi': 'Guru',
        'Gujarati': 'Gujr',
        'Oriya': 'Orya',
        'Tamil': 'Taml',
        'Telugu': 'Telugu',
        'Kannada': 'Knda',
        'Malayalam': 'Mlym',
        'Sinhala': 'Sinh',
        'Thai': 'Thai',
        'Lao': 'Laoo',
        'Tibetan': 'Tibt',
        'Myanmar': 'Mymr',
        'Georgian': 'Geor',
        'Khmer': 'Khmr',
        'SimplifiedChinese': 'Hans',
        'TraditionalChinese': 'Hant',
        'Japanese': 'Jpan',
        'Korean': 'Kore',
        #'Vietnamese': 'Latn',
        #'Symbol': 'Zsym',
        ##'Other': 'Zyyy', # Alias for Symbol in Qt
        'Ogham': 'Ogam',
        'Runic': 'Runr',
        'Nko': 'Nkoo',
    }

    languages = {
        'Vietnamese': 'vi',

        #'Latin': many
        'Greek': 'el',
        'Armenian': 'hy',
        'Hebrew': ['he', 'yi', 'lad', 'mxi'],
        'Arabic': ['ar', 'fa', 'ms', 'ur'],
        #'Syriac': many
        'Thaana': 'dv',
        #'Devanagari': many
        'Bengali': ['bn'], # ...and Sanskrit 'sa'
        'Gurmukhi': ['pa', 'sd'],
        'Gujarati': ['gu', 'kfr'], # and many more
        'Oriya': ['or'], # ...and many regional
        'Tamil': ['ta'], # ...and many regional
        'Telugu': ['te'],
        'Kannada': ['kn'],
        'Malayalam': 'ml',
        'Sinhala': 'si',
        'Thai': ['th', 'sou'], # ...and many more
        'Lao': ['lo'], # ...and many regional
        'Tibetan': ['bo', 'dz', 'sip', 'lbj', 'zau', 'jul'],
        'Myanmar': 'my',
        'Georgian': 'ka',
        'Khmer': 'km',
        'SimplifiedChinese': ['zh-CN', 'zh-SG'],
        'TraditionalChinese': ['zh-TW', 'zh-HK', 'zh-MO'],
        'Japanese': ['ja', 'ain'],
        'Korean': ['ko', 'jje', 'cia'],
        #'Ogham': 'Ogam',
        #'Runic': 'Runr',
        'Nko': ['nqo', 'bm', 'dyu',
                'mku', 'emk', 'msc', 'mzj', 'jod', 'jud',
                'kfo', 'kga', 'mxx'],
    }

    # Ensure some consistency for now
    for key, value in list(languages.items()):
        if not isinstance(value, list):
            languages[key] = [value]

    result = dict(
        writingSystemNameToScript=scripts,
        writingSystemNameToLanguages=languages,
        writingSystemToScript={},
        writingSystemToLanguages={},

        scriptToWritingSystemName={},
        languageToWritingSystemName={},
        scriptToWritingSystem={},
        languageToWritingSystem={},
    )

    for wsName, script in list(scripts.items()):

        result['scriptToWritingSystemName'][script] = wsName

        ws = getattr(QtGui.QFontDatabase, key, None)
        if (ws is None or
            not isinstance(ws, QtGui.QFontDatabase.WritingSystem)):
                continue

        result['writingSystemToScript'][ws] = script
        result['scriptToWritingSystem'][script] = ws

    for wsName, langs in list(languages.items()):
        for lang in langs:
            result['languageToWritingSystemName'][lang] = wsName

        ws = getattr(QtGui.QFontDatabase, key, None)
        if (ws is None or
            not isinstance(ws, QtGui.QFontDatabase.WritingSystem)):
                continue

        result['writingSystemToLanguages'][ws] = langs
        for lang in langs:
            result['languageToWritingSystem'][lang] = ws

    _writingSystemAliases = result

    return result



# -*- coding: utf-8 -*-
#
#    TypeAtlas Character Set Functions
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#
# coding: utf-8


"""Utilities for loading non-Unicode encoding / character set data, mostly
used for the character map in legacy mode.
"""

import sys
import copy
import pkgutil
import importlib
import unicodedata
import encodings as encodings_mod
import binascii
from encodings import aliases as enc_aliases
from operator import itemgetter

from collections import namedtuple, OrderedDict, Counter
from collections.abc import Set, Callable

from typeatlas import charinfo, blockmath
from typeatlas.util import OrderedSet, N_, warnmsgf, noticemsgf, generic_type
from typeatlas.util import EmptyMapping, EmptySequence

MappingOf = generic_type('Mapping')
IterableOf = generic_type('Iterable')
IteratorOf = generic_type('Iterator')
SetOf = generic_type('AbstractSet')
SequenceOf = generic_type('Sequence')
TupleOf = generic_type('Tuple')
Union = generic_type('Union')
Optional = generic_type('Optional')


UNICODE = EmptySequence()
UNICODE_SCRIPTS = EmptyMapping()

CONTROL_CHAR_COUNT = 31
DELETE_CHAR = 127

EBCDIC_CODEPAGES = frozenset([
    '037', '273', '277', '278', '280',
    '284', '285', '297', '500', '871', '1047',
    '1140', '1141', '1142', '1143', '1144',
    '1145', '1146', '1147', '1148', '1149', '924'])

DOS_CODEPAGES = frozenset([
    '100', '111', '112', '113', '151', '152', '161', '162', '163', '164',
    '165', '166', '210', '220', '301', '437', '449', '489', '620', '667',
    '668', '707', '708', '709', '710', '711', '714', '715', '720', '721',
    '737', '768', '770', '771', '772', '773', '774', '775', '776', '777',
    '778', '790', '850', '851', '852', '853', '854', '855', '872', '856',
    '857', '858', '859', '860', '861', '862', '863', '864', '17248', '865',
    '866', '808', '867', '868', '869', '874', '1161', '1162', '876', '877',
    '878', '881', '882', '883', '884', '885', '891', '895', '896', '897',
    '898', '899', '900', '903', '904', '906', '907', '909', '910', '911',
    '926', '927', '928', '929', '932', '934', '936', '938', '941', '942',
    '943', '944', '946', '947', '948', '949', '950', '1370', '951', '966',
    '991', '1034', '1039', '1040', '1041', '1042', '1043', '1044', '1046',
    '1086', '1088', '1092', '1093', '1098', '1108', '1109', '1114', '1115',
    '1116', '1117', '1118', '1119', '1125', '848', '1126', '1127', '1131',
    '849', '1139', '1167', '1168', '1300', '1351', '1361', '1362', '1363',
    '1372', '1373', '1374', '1375', '1380', '1381', '1385', '1386', '1391',
    '1392', '1393', '1394'])
    
WINDOWS_ENCODINGS = set()
WINDOWS_ENCODINGS.update(val
                         for key, val in enc_aliases.aliases.items()
                            if key.startswith('windows'))
WINDOWS_ENCODINGS.update(key
                         for key, val in enc_aliases.aliases.items()
                            if val in WINDOWS_ENCODINGS)


# KOI* are DOS charsets, but were also used after DOS, so don't include
# them here. The following encodings are DOS encodings that are missing
# (i.e. we don't have them in encodings.py yet):
# * Iran System
# * CWI-2/HUCWI/cp-hu for Hungarian (slightly changed cp437)
# * Kamenický/KEYBCS2 for Czech (based on cp437) (unoffically cp895)
# * Mazovia for Polish (called 667 and 991 in some DOS versions, 
#                       though 991 is ambiguous)
# * MIK for Bulgarian
# are missing, perhaps altogether.
    

encodings = OrderedDict()


def add_encoding(name: str, *args, **kwargs):
    """Add a known encoding, creating the Encoding object and
    registring it with the modules's encodings dictionary."""
    encodings[name] = Encoding(name, *args, **kwargs)


morse_codes = OrderedDict()


def add_morse_code(name: str, *args, **kwargs):
    """Add a known morse code or similar ancient encoding, creating
    the MorseCode object and registring it with the module's
    morse_codes dictionary."""
    morse_codes[name] = MorseCode(name, *args, **kwargs)


exclude = set(['aliases', 'charmap', 'base64_codec',
               'zlib_codec', 'uu_codec', 'bz2_codec',
               'raw_unicode_escape', 'unicode_escape',
               'unicode_internal', 'rot_13', 'punycode',
               'idna', 'undefined'])


dos_graphical_context = dict(enumerate('☺☻♥♦♣♠•◘○◙♂♀♪♫☼►◄↕‼¶§▬↨↑↓→←∟↔▲▼', 1))
assert len(dos_graphical_context) == CONTROL_CHAR_COUNT
dos_graphical_context[DELETE_CHAR] = '⌂'

uhc_graphical_context = dict(enumerate('┌┐└┘│─•◘○◙♂♀♪♫☼┼◄↕‼¶┴┬┤↑├→←∟↔▲▼', 1))
assert len(uhc_graphical_context) == CONTROL_CHAR_COUNT
uhc_graphical_context[DELETE_CHAR] = '⌂'

shift_jis_overrides = {b'\\': '¥',
                       b'~': '‾'}
uhc_overrides = {b'\\': '₩'}


encoding_overrides = {
    'shift_jis': shift_jis_overrides,
    'euc_jp': shift_jis_overrides,
    'uhc': uhc_overrides,
    'euc_kr': uhc_overrides,
    'cp949': uhc_overrides,
}


graphical_overrides = {
    'default': {bytes([code]): char
                for code, char in dos_graphical_context.items()},
    'cp949': {bytes([code]): char
              for code, char in uhc_graphical_context.items()},
    'uhc': {bytes([code]): char
            for code, char in uhc_graphical_context.items()},
}


class InterpretationBase(object):

    """This is a base class for interpretations of characters in
    character encodings (for codes whose translations to unicode are
    ambigiuous).

    All subclasses have label, icon, and optional help.

    This used in both the implementation of interpretations, and
    interpretation groups. Use them."""

    def __init__(self, label: str, icon: str=None, help: str=None):

        self.label = label
        self.icon = icon
        self.help = help

    @classmethod
    def make(cls, *args, **kwargs) -> 'InterpretationBase':
        """Create a new interpretation (or group) and register it with
        as a known interpretation (group)."""
        result = cls(*args, **kwargs)
        cls.available.append(result)
        return result

    def __repr__(self):
        return '<%s %r at 0x%x>' % (type(self).__name__, self.label, id(self))


BytedataOverrideType = MappingOf[bytes, str]
EncodingOverridesType = MappingOf[Optional[str], BytedataOverrideType]


class Interpretation(InterpretationBase, Set):

    """Define a different interpretation of the characters depending
    on context. They also interpret a Set interface, so they can be
    used interchangeably with a set of interpretions (InterpretationSet).

    Each Interpretation can be - and should be - in a given group, can be
    the default interpretation in a group if configured, and have a
    dictionary of overrides for given encodings, or general ones
    (signified with None).
    """

    available = []

    def __init__(self, *args,
                       group: 'InterpretationGroup'=None,
                       default: bool=False,
                       overrides: EncodingOverridesType=(),
                       condition: Callable=None,
                       **kwargs):
        super().__init__(*args, **kwargs)
        self.group = group
        self.default = default
        self.overrides = dict(overrides)
        self.condition = condition

    @classmethod
    def make(cls, *args, group: 'InterpretationGroup'=None,
                  **kwargs) -> 'Interpretation':
        """Create a new interpretation and register it with
        as a known interpretation."""
        result = super().make(*args, group=group, **kwargs)
        if group is not None:
            group.interpretations.add(result)
        return result

    def bytedata_remap(self, encoding: str) -> MappingOf[bytes, str]:
        """Get remappings of bytedata to unicode characters for a given encoding."""
        if encoding.unicode:
            return {}
        if encoding.name in self.overrides:
            return self.overrides[encoding.name]
        elif self.condition is None or self.condition(encoding):
            return self.overrides.get('default') or {}
        else:
            return {}

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        self is other

    def __ne__(self, other):
        self is not other

    def __contains__(self, item):
        return self is item

    def __iter__(self):
        yield self

    def __len__(self):
        return 1

    @classmethod
    def _from_iterable(cls, iterable):
        return InterpretationSet(iterable)


class InterpretationSet(Set):

    """A set of enabled interpretations that are used together to re-interpret
    characters."""

    def __init__(self, interpretations: IterableOf[Interpretation]):
        self.interpretations = OrderedSet(interpretations)

    def bytedata_remap(self, encoding: str) -> MappingOf[bytes, str]:
        """Get remappings of bytedata to unicode characters for a given encoding,
        getting it from all encodings in the set."""
        result = {}
        for interpretation in self.interpretations:
            result.update(interpretation.bytedata_remap(encoding))
        return result

    def __contains__(self, item):
        return item in self.interpretations

    def __iter__(self):
        return iter(self.interpretations)

    def __len__(self):
        return len(self.interpretations)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.interpretations)


class InterpretationGroup(InterpretationBase, Set):

    """A group of interpretations (usually mutually exclusive) that can be
    selected. The mutual exclusivety is specified using the 'exclusive'
    argument and attribute."""

    available = []

    def __init__(self, *args,
                        interpretations: IterableOf[Interpretation]=(),
                        exclusive: bool=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.interpretations = OrderedSet(interpretations)
        self.exclusive = bool(exclusive)

    def __contains__(self, item):
        return item in self.interpretations

    def __iter__(self):
        return iter(self.interpretations)

    def __len__(self):
        return len(self.interpretations)

    @classmethod
    def _from_iterable(cls, iterable):
        return cls(label='', interpretations=iterable)


InterpretationLike = Union[InterpretationSet, Interpretation]


CONTROL_CHAR_INTERPRETATIONS = InterpretationGroup.make(
            N_('Control characters'), icon='character-control',
            help=N_('Intereptations of the 7-bit ASCII control characters'),
            exclusive=True)

AMBIGIOUS_CHAR_INTERPRETATIONS = InterpretationGroup.make(
            N_('Ambiguous characters'), icon='character-ambiguous',
            help=N_('Intereptations of the 7-bit ASCII control characters'),
            exclusive=True)


ASCII_CONTROL_INTERPRETATION = Interpretation.make(
            N_('ASCII'), icon='character-control',
            help=N_('ASCII control code interpretation of leading CP437 range'),
            group=CONTROL_CHAR_INTERPRETATIONS, default=True)

GRAPHICAL_INTERPRETATION = Interpretation.make(
            N_('Graphical'),
            #icon='character-symbol',
            icon='character-ascii-graphical',
            help=N_('Graphical interpretation of leading CP437 range'),
            group=CONTROL_CHAR_INTERPRETATIONS, default=False,
            overrides=graphical_overrides,
            condition=lambda enc: enc.name.startswith('cp') and
                                  enc.name[2:] in DOS_CODEPAGES)

LOCALE_INTERPRETATION = Interpretation.make(
            N_('Locale interpretation'),
            help=N_('The correct interpretation in the locale for '
                    'ambiguous Shift-JIS or UHC characters'),
            group=AMBIGIOUS_CHAR_INTERPRETATIONS, default=True,
            overrides=encoding_overrides)

ASCII_COMPAT_INTERPRETATION = Interpretation.make(
            N_('ASCII compatibility'),
            help=N_('ASCII equivalent interpretation of '
                    'ambiguous Shift-JIS or UHC characters'),
            group=AMBIGIOUS_CHAR_INTERPRETATIONS, default=False)


class Encoding(object):

    """Information about a given encoding.

    The information can contain ISO 15924 codes of supported scripts,
    a list of characters supported, whether it is a multibyte encoding,
    whether it is unicode, whether it is fixed-width - and what is the
    charsize if true, what maximum bits can it use, and what mapping
    of codepoints to bytes it might use (if applicable).
    """

    morse = False

    def __init__(self, name: str,
                       scripts: MappingOf[str, int]=(),
                       characters: SequenceOf[str]=None,
                       multibyte: bool=False,
                       unicode: bool=False, fixed: bool=True,
                       charsize: int=None, maxbits: int=None,
                       codepoint_bytes: MappingOf[int, bytes]=None):

        self.name = name
        self.scripts = scripts
        self.characters = characters
        self.multibyte = multibyte
        self.unicode = unicode
        self.fixed = fixed
        self.charsize = charsize
        self.maxbits = maxbits

        self._canonical_codepoint_bytes = codepoint_bytes
        self._remapped_codepoint_bytes = None

        self.current_interpretation = None
        self.bytedata_remap = self.codepoint_remap = self.char_remap = None

    def interpretation(self, interpretation: InterpretationLike) -> 'Encoding':
        """Get an alternative interpretation of the encoding."""
        result = copy.copy(self)
        result.current_interpretation = interpretation

        if interpretation is not None:
            result.bytedata_remap = interpretation.bytedata_remap(self)
        else:
            result.bytedata_remap = None

        if not result.bytedata_remap:
            result.bytedata_remap = None
            result.codepoint_remap = None
            result.char_remap = None
            return result

        result.codepoint_remap = {ord(char): bytedata
                                  for bytedata, char
                                        in result.bytedata_remap.items()}
        result.char_remap = {char: bytedata
                             for bytedata, char
                                    in result.bytedata_remap.items()}

        return result

    def codepoint_bytes(self, chardb: charinfo.CharacterDatabase=None, *,
                              cache: bool=False) -> MappingOf[int, bytes]:
        """Get a mapping of codepoint code to bytedata, respecting the current
        interpretation."""
        result = self._remapped_codepoint_bytes
        if result is not None:
            return result

        bytedata_remap = self.bytedata_remap

        canonical_result = self.canonical_codepoint_bytes(chardb, cache=cache)
        if bytedata_remap is None:
            return canonical_result

        result = {}

        for codepoint, bytedata in canonical_result.items():
            if bytedata in bytedata_remap:
                codepoint = ord(bytedata_remap[bytedata])
            result[codepoint] = bytedata

        if cache:
            self._remapped_codepoint_bytes = result

        return result

    def canonical_codepoint_bytes(self,
                                  chardb: charinfo.CharacterDatabase=None, *,
                                  cache: bool=False) -> MappingOf[int, bytes]:
        """Get a mapping of codepoint code to bytedata without respect to
        current interpretation."""
        result = self._canonical_codepoint_bytes
        if result is not None:
            return result
        if not self.characters:
            # FIXME: No point caching in a copy of class that we throw away
            return self.complete(chardb).canonical_codepoint_bytes(chardb, cache=cache)
        result = {}
        for char in self.characters:
            bytedata = char.encode(self.name)
            result[ord(char)] = bytedata

        if cache:
            self._canonical_codepoint_bytes = result

        return result

    def complete(self, chardb: charinfo.CharacterDatabase=None) -> 'Encoding':
        """Get a complete version of the encoding, with all the information,
        particularly about the supported characters, filled by
        get_encoding_slow()."""

        if self.characters:
            return self
        result = copy.copy(self)
        #result.characters = ''.join(map(itemgetter(1),
        #                                get_characters_slow(self.name)))
        complete = get_encoding_slow(self.name, chardb)
        result.characters = complete.characters
        result.codepoint_bytes = complete.codepoint_bytes
        result.multibyte = complete.multibyte
        result.fixed = complete.fixed
        result.charsize = complete.charsize
        result.maxbits = complete.maxbits
        result._canonical_codepoint_bytes = complete._canonical_codepoint_bytes
        result._remapped_codepoint_bytes = None
        return result

    def get_byte_data(self, code: int) -> bytes:
        """Get the byte data for a given code point."""
        if self.codepoint_remap is not None and code in self.codepoint_remap:
            return self.codepoint_remap[code]
        return chr(code).encode(self.name)

    def get_hex(self, code: int) -> str:
        """Get the hex-encoded version of the byte data for a given code point."""
        return str(binascii.hexlify(self.get_byte_data(code)))

    def get_integer(self, code: int) -> int:
        """Get a useless integirifed version of the byte data for a given code
        point. That's only sensible for fixed-length encodings or encodings
        that contain no null bytes."""
        result = 0
        for c in self.get_byte_data(code):
            result = result * 256 + c
        return result

    def label(self, translate: Callable=N_) -> str:
        """Get the label for the given encoding.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is."""

        _ = translate

        name = self.name

        label = name.upper().replace('_', '-')

        if name.startswith('cp'):
            codepage = name[2:]
            if codepage in DOS_CODEPAGES:
                return _('DOS code page %s') % (name[2:], )
            if codepage in EBCDIC_CODEPAGES:
                return _('EBCDIC code page %s') % (name[2:], )

        if name in WINDOWS_ENCODINGS and name.startswith('cp'):
            return 'Windows ' + name[2:]

        return label

    common_scripts = set(['Zxxx', 'Zyyy', 'Zzzz'])

    def category(self) -> Optional[str]:
        """Return a category to put the encoding in for display.

        One possible interpretation is that it detects the encoding
        script(s) and returns one of them."""
        if self.scripts is UNICODE_SCRIPTS:
            return None

        scripts = [s for s in self.scripts
                   if s not in self.common_scripts]
        if not scripts:
            return None
        return max(scripts, key=self.scripts.get)

    def category_label(self, translate: Callable=N_) -> str:
        """Return a category label to put the encoding in for display,
        optionally translated.

        One possible interpretation is that it detects the encoding
        script(s) and returns one of them.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is."""
        _ = translate

        if self.unicode:
            return _('Unicode')

        category = self.category()
        if category is None:
            return _("Other")

        from typeatlas import langutil
        langdb = langutil.LanguageDatabase.get_instance(populated=True)
        return langdb.script_name(category)

    def get_icon(self) -> Optional[str]:
        """Return the name of an icon for the given encoding. This will either be
        None, or a file provided with typeatlas without the extension."""
        name = self.name

        if name.startswith('cp'):
            codepage = name[2:]
            if codepage in DOS_CODEPAGES:
                return 'platform-dos'
            if codepage in EBCDIC_CODEPAGES:
                return 'platform-oldiron'
            if name in WINDOWS_ENCODINGS:
                return 'platform-redmond'

        elif name.startswith('mac'):
            return 'platform-cupertino'

        elif name.startswith('iso'):
            return 'standard'

        elif name in WINDOWS_ENCODINGS or name.startswith('windows'):
            return 'platform-redmond'

    def __repr__(self):
        return '<%s %r at 0x%x>' % (type(self).__name__, self.name, id(self))


ONE_TO_ONE_REMAP = 0
MULTICHAR_REMAP = 1


class MorseCode(object):

    """Class that implements the same interface as Encoding, but for morse
    code and similar encodings.

    A code is encoded using an alphabet, which maps unicode character to
    code (if lenght_coded is True, this can use 'dashes' of multiple lengths).

    International morse codes can instead map provide those as remap base, and
    map international unicode characters to latin unicode characters in the base.
    If remap_mode is MULTICHAR_REMAP, then one character can be translated to
    multiple.

    The final mapping (from character to telegraph code) is stored
    in self.alpha_codes.

    If needle_coded is True, then the code is not morse-code-like, but a needle
    telegraph code. This is not supported at the time of documenting the code.
    """

    morse = True

    def __init__(self, name: str,
                       alphabet: MappingOf[str, str],
                       alphabet_remap_base: MappingOf[str, str]=None,
                       remap_mode=ONE_TO_ONE_REMAP,
                       length_coded: bool=False,
                       needle_coded: bool=False,
                       scripts: MappingOf[str, int]=(),
                       characters: SequenceOf[str]=None,

                       # Unused
                       charsize: int=None, maxbits: int=None,
                       codepoint_bytes: MappingOf[int, bytes]=None):

        self.name = name

        if alphabet_remap_base is None:
            self.alpha_codes = alphabet
            self.alpha_remap = {}
            self.base_codes = alphabet

        elif remap_mode == ONE_TO_ONE_REMAP:
            self.alpha_codes = {alpha: alphabet_remap_base[char]
                                for alpha, char in alphabet.items()}
            self.alpha_remap = alphabet
            self.base_codes = alphabet_remap_base

        elif remap_mode == MULTICHAR_REMAP:
            self.alpha_codes = {alpha: ' '.join(alphabet_remap_base[char]
                                                for char in chars)
                                for alpha, chars in alphabet.items()}
            self.alpha_remap = alphabet
            self.base_codes = alphabet_remap_base

        self.remap_mode = remap_mode
        self.length_coded = length_coded
        self.needle_coded = needle_coded

        self.scripts = scripts
        self.characters = characters
        self.multibyte = True
        self.unicode = False
        self.fixed = False
        self.charsize = None
        self.maxbits = maxbits
        self._canonical_codepoint_bytes = codepoint_bytes

    def alpha_telegraph_codes(self, pretty: bool=True) -> MappingOf[str, str]:
        """Return mapping of unicode characters to telegraph code for display.

        If pretty=False is true, the raw (not the pretty) version is
        returned."""
        if not pretty or self.needle_coded:
            return dict(self.alpha_codes)

        remaps = {
            ord('.'): '▄',
            ord('-'): '▄' if self.length_coded else '▄▄',
        }

        return {char: text.translate(remaps)
                for char, text in self.alpha_codes.items()
                if len(char) == 1}

    def codepoint_telegraph_codes(self, *args, **kwargs) -> MappingOf[int, str]:
        """Return mapping of unicode codepoints to telegraph code for display.

        If pretty=False is true, the raw (not the pretty) version is
        returned."""

        return {ord(char): text
                for char, text
                    in self.alpha_telegraph_codes(*args, **kwargs).items()
                if len(char) == 1}


    def codepoint_bytes(self, chardb: charinfo.CharacterDatabase=None, *,
                              cache: False=False) -> MappingOf[int, bytes]:
        if self._canonical_codepoint_bytes is None:
            self._canonical_codepoint_bytes = {
                ord(char): (morse + '  ').encode('ascii')
                for char, morse in self.alpha_codes.items()
            }

        return self._canonical_codepoint_bytes

    def interpretation(self, interpretation: InterpretationLike) -> 'Encoding':
        return self

    def complete(self, chardb: charinfo.CharacterDatabase=None) -> 'MorseCode':
        return self

    def get_byte_data(self, code: int) -> bytes:
        return (self.alpha_codes.get(chr(code)) + '  ').encode('ascii')

    def get_hex(self, code: int) -> bytes:
        return str(binascii.hexlify(self.get_byte_data(code)))

    def get_integer(self, code: int) -> int:
        result = 0
        for c in self.get_byte_data(code):
            result = result * 256 + c
        return result

    def label(self, translate: Callable=N_) -> str:
        """Get the label for the given telegraph code.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is.
        """
        _ = translate

        return _(self.name)

    common_scripts = set(['Zxxx', 'Zyyy', 'Zzzz'])

    def category(self) -> Optional[str]:
        """Return a category to put the telegraph code in for display."""
        scripts = [s for s in self.scripts
                   if s not in self.common_scripts]
        if not scripts:
            return None
        return max(scripts, key=self.scripts.get)

    def category_label(self, translate: Callable=N_) -> str:
        """Return a category label to put the telegraph code in for display,
        optionally translated.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is."""
        _ = translate

        category = self.category()
        if category is None:
            return _("Other")

        from typeatlas import langutil
        langdb = langutil.LanguageDatabase.get_instance(populated=True)
        return langdb.script_name(category)

    def get_icon(self) -> Optional[str]:
        """Return an icon for the given telegraph code."""
        #return 'morse'
        return None


EncodingLike = Union[Encoding, MorseCode]


def charblocks(blocks: IterableOf[blockmath.BlockLike]) -> str:
    """Convert an iterable of character blocks to a string of characters."""
    return ''.join(map(chr, blockmath.iterblocks(blocks)))


def get_encodings() -> IterableOf[Encoding]:
    """Return an iterable of all encodings."""

    result = set(name
                 for loader, name, ispkg
                    in pkgutil.iter_modules(encodings_mod.__path__)
                 if not ispkg)

    result -= exclude
    result -= set(enc_aliases.aliases)

    return result


def get_characters_slow(charset: str) -> IteratorOf[TupleOf[int, str, bytes]]:
    """Slow function to find the characters in an encoding.

    Returns tuple of codepoint, character and encodedbytedata."""

    try:
        module = importlib.import_module('encodings.utf_8')
    except ImportError:
        pass
    else:
        if hasattr(module, 'decoding_table'):
            result = []
            for i, char in sorted(enumerate(module.decoding_table),
                                  key=itemgetter(1)):
                if char == '\ufffe':
                    continue
                assert char.encode(charset) == bytes([i])
                yield ord(char),  char, char.encode(charset)
            return

    overrides = encoding_overrides.get(charset, {})
    seen = {}
    result = OrderedDict()

    for code in charinfo.all_character_codes():
        char = chr(code)
        try:
            bytedata = char.encode(charset)
        except ValueError:
            continue

        if bytedata in seen:

            action = 'kept'

            # FIXME: Why are we doing this?
            if bytedata in overrides:
                replace = overrides[bytedata] == char
            else:
                replace = bytedata.decode(charset) == char

            if replace:
                action = 'replaced'
                del result[char]
                result[char] = code, char, bytedata

            if (unicodedata.normalize('NFKC', char) ==
                unicodedata.normalize('NFKC', seen[bytedata])):
                showmsgf = noticemsgf
            else:
                showmsgf = warnmsgf

            showmsgf("%r encodes both %r and %r, %s the former",
                     bytedata, seen[bytedata], char, action)

            continue

        seen[bytedata] = char

        result[char] = code, char, bytedata

    yield from result.values()


def get_encoding(name: str) -> Encoding:
    """Get encoding by name."""
    name = encodings_mod.normalize_encoding(name)
    return encodings.get(name)


def get_encoding_slow(name,
                      chardb: charinfo.CharacterDatabase=None) -> Encoding:
    """Slow function to get encoding info."""

    if chardb is None:
        chardb = charinfo.CharacterDatabase.get_instance(populated=True)

    codepoints = OrderedDict()
    characters = OrderedSet()

    charsizes = set()

    clean_7bit = True

    for code, char, bytedata in get_characters_slow(name):

        bytelen = len(bytedata)

        if bytelen == 1 and 0b10000000 & ord(bytedata):
            clean_7bit = False

        codepoints[code] = bytedata
        charsizes.add(bytelen)
        characters.add(char)

    fixed = len(charsizes) == 1
    multibyte = charsizes != {1}

    if fixed:
        charsize = next(iter(charsizes))
    else:
        charsize = None
    scripts = Counter(chardb.find_scripts(codepoints))
    scripts = {chardb.get_value_alias('sc', script): count
               for script, count in scripts.items()}

    maxbits = max(charsizes) * 8
    if maxbits == 8 and clean_7bit:
        maxbits = 7

    return Encoding(name, scripts,
                    characters=''.join(characters),
                    multibyte=multibyte, fixed=fixed, charsize=charsize,
                    maxbits=maxbits, codepoint_bytes=codepoints)


def print_encoding_info():
    """Print info about all encodings"""

    charconsts = {}


    utf8 = get_encoding_slow('utf8')
    for charset in sorted(get_encodings()):
        try:
            encoding = get_encoding_slow(charset)
        except LookupError:
            continue

        unicode = False

        scriptrepr = repr(encoding.scripts)

        characters = ''.join(sorted(encoding.characters))
        if characters == utf8.characters:

            charconst = 'UNICODE'
            unicode = True
            scriptrepr = 'UNICODE_SCRIPTS'

        elif characters in charconsts:
            charconst = charconsts[characters]

        elif len(characters) > 1024:
            charconst = 'None'

        else:
            blocks = blockmath.toblocks(map(ord, characters))
            blockcode = 'charblocks([%s])' % (
                            ', '.join('(0x%x, 0x%x)' % block
                                      for block in blocks))

            charconst = 'CHARS_' + charset.upper()

            if len(characters) < 512:
                print('# %s = %r\n\n' % (charconst, characters))

            charconsts[characters] = charconst
            print('%s = %s\n\n' % (charconst, blockcode))

        print('''
add_encoding({name!r}, characters={characters},
             scripts={scripts},
             multibyte={multibyte!r}, fixed={fixed!r}, unicode={unicode!r},
             charsize={charsize!r}, maxbits={maxbits!r})
'''.format(name=encoding.name, scripts=scriptrepr,
           characters=charconst,
           multibyte=encoding.multibyte, fixed=encoding.fixed,
           unicode=unicode, charsize=encoding.charsize, 
           maxbits=encoding.maxbits))


def load_encodings():
    """Load the encodings. This is automatically called."""
    from typeatlas.data import encodings as encoding_infos
    from typeatlas.data import morse as morse_infos
    encoding_infos, morse_infos

    for encoding in get_encodings():
        if encoding not in encodings:
            try:
                'test'.encode(encoding, 'replace')
            except (ValueError, TypeError, LookupError):
                pass
            else:
                encodings[encoding] = Encoding(encoding)


load_encodings()

if __name__ == '__main__':
    print_encoding_info()
